

__all__ = ['apply_transforms',
           'apply_transforms_to_points']

import os

import ants
from ants.internal import get_lib_fn, process_arguments


def apply_transforms(fixed, moving, transformlist,
                     interpolator='linear', imagetype=0,
                     whichtoinvert=None, compose=None,
                     defaultvalue=0, singleprecision=False, verbose=False, **kwargs):
    """
    Apply a transform list to map an image from one domain to another.
    In image registration, one computes mappings between (usually) pairs
    of images. These transforms are often a sequence of increasingly
    complex maps, e.g. from translation, to rigid, to affine to deformation.
    The list of such transforms is passed to this function to interpolate one
    image domain into the next image domain, as below. The order matters
    strongly and the user is advised to familiarize with the standards
    established in examples.

    ANTsR function: `antsApplyTransforms`

    Arguments
    ---------
    fixed : ANTsImage
        fixed image defining domain into which the moving image is transformed. The output will
        have the same pixel type as this image.

    moving : AntsImage
        moving image to be mapped to fixed space.

    transformlist : list of strings
        list of transforms generated by ants.registration where each transform is a filename.

    interpolator : string
        Choice of interpolator. Supports partial matching.
            linear
            nearestNeighbor
            multiLabel for label images (deprecated, prefer genericLabel)
            gaussian
            bSpline
            cosineWindowedSinc
            welchWindowedSinc
            hammingWindowedSinc
            lanczosWindowedSinc
            genericLabel use this for label images

    imagetype : integer
        choose 0/1/2/3 mapping to scalar/vector/tensor/time-series

    whichtoinvert : list of booleans (optional)
        Must be same length as transformlist.
        whichtoinvert[i] is True if transformlist[i] is a matrix,
        and the matrix should be inverted. If transformlist[i] is a
        warp field, whichtoinvert[i] must be False.
        If the transform list is a matrix followed by a warp field,
        whichtoinvert defaults to (True,False). Otherwise it defaults
        to [False]*len(transformlist)).

    compose : string (optional)
        if it is a string pointing to a valid file location,
        this will force the function to return a composite transformation filename.

    defaultvalue : scalar
        Default voxel value for mappings outside the image domain.

    singleprecision : boolean
        if True, use float32 for computations. This is useful for reducing memory
        usage for large datasets, at the cost of precision.

    verbose : boolean
        print command and run verbose application of transform.

    kwargs : keyword arguments
        extra parameters

    Returns
    -------
    ANTsImage or string (transformation filename)

    Example
    -------
    >>> import ants
    >>> fixed = ants.image_read( ants.get_ants_data('r16') )
    >>> moving = ants.image_read( ants.get_ants_data('r64') )
    >>> fixed = ants.resample_image(fixed, (64,64), 1, 0)
    >>> moving = ants.resample_image(moving, (64,64), 1, 0)
    >>> mytx = ants.registration(fixed=fixed , moving=moving ,
                                 type_of_transform = 'SyN' )
    >>> mywarpedimage = ants.apply_transforms( fixed=fixed, moving=moving,
                                               transformlist=mytx['fwdtransforms'] )
    """

    if not isinstance(transformlist, (tuple, list)) and (transformlist is not None):
        transformlist = [transformlist]

    accepted_interpolators = {"linear", "nearestNeighbor", "multiLabel", "gaussian",
                        "bSpline", "cosineWindowedSinc", "welchWindowedSinc",
                        "hammingWindowedSinc", "lanczosWindowedSinc", "genericLabel"}

    if interpolator not in accepted_interpolators:
        raise ValueError('interpolator not supported - see %s' % accepted_interpolators)

    args = [fixed, moving, transformlist, interpolator]

    output_pixel_type = 'float' if singleprecision else 'double'

    if not isinstance(fixed, str):
        if ants.is_image(fixed) and ants.is_image(moving):
            for tl_path in transformlist:
                if not os.path.exists(tl_path):
                    raise Exception('Transform %s does not exist' % tl_path)

            inpixeltype = fixed.pixeltype
            fixed = fixed.clone(output_pixel_type)
            moving = moving.clone(output_pixel_type)
            warpedmovout = moving.clone(output_pixel_type)
            f = fixed
            m = moving
            if (moving.dimension == 4) and (fixed.dimension == 3) and (imagetype == 0):
                raise Exception('Set imagetype 3 to transform time series images.')

            wmo = warpedmovout
            mytx = []
            if whichtoinvert is None or (isinstance(whichtoinvert, (tuple,list)) and (sum([w is not None for w in whichtoinvert])==0)):
                if (len(transformlist) == 2) and ('.mat' in transformlist[0]) and ('.mat' not in transformlist[1]):
                    whichtoinvert = (True, False)
                else:
                    whichtoinvert = tuple([False]*len(transformlist))

            if len(whichtoinvert) != len(transformlist):
                raise ValueError('Transform list and inversion list must be the same length')

            for i in range(len(transformlist)):
                ismat = False
                if '.mat' in transformlist[i]:
                    ismat = True
                if whichtoinvert[i] and (not ismat):
                    raise ValueError('Cannot invert transform %i (%s) because it is not a matrix' % (i, transformlist[i]))
                if whichtoinvert[i]:
                    mytx = mytx + ['-t', '[%s,1]' % (transformlist[i])]
                else:
                    mytx = mytx + ['-t', transformlist[i]]

            if compose is None:
                args = ['-d', fixed.dimension,
                        '-i', m,
                        '-o', wmo,
                        '-r', f,
                        '-n', interpolator]
                args = args + mytx
            if compose:
                tfn = '%scomptx.nii.gz' % compose if not compose.endswith('.h5') else compose
            else:
                tfn = 'NA'
            if compose is not None:
                mycompo = '[%s,1]' % tfn
                args = ['-d', fixed.dimension,
                        '-i', m,
                        '-o', mycompo,
                        '-r', f,
                        '-n', interpolator]
                args = args + mytx

            myargs = process_arguments(args)

            myverb = int(verbose)
            if verbose:
                print(myargs)

            processed_args = myargs + ['-z', str(1), '-v', str(myverb), '--float', str(int(singleprecision)), '-e', str(imagetype), '-f', str(defaultvalue)]
            libfn = get_lib_fn('antsApplyTransforms')
            libfn(processed_args)

            if compose is None:
                return warpedmovout.clone(inpixeltype)
            else:
                if os.path.exists(tfn):
                    return tfn
                else:
                    return None

        else:
            return 1
    else:
        args = args + ['-z', str(1), '--float', str(int(singleprecision)), '-e', imagetype, '-f', defaultvalue]
        processed_args = process_arguments(args)
        libfn = get_lib_fn('antsApplyTransforms')
        libfn(processed_args)






def apply_transforms_to_points( dim, points, transformlist,
                     whichtoinvert=None, verbose=False ):
    """
     Apply a transform list to map a pointset from one domain to
     another. In registration, one computes mappings between pairs of
     domains. These transforms are often a sequence of increasingly
     complex maps, e.g. from translation, to rigid, to affine to
     deformation.  The list of such transforms is passed to this
     function to interpolate one image domain into the next image
     domain, as below.  The order matters strongly and the user is
     advised to familiarize with the standards established in examples.
     Importantly, point mapping goes the opposite direction of image
     mapping, for both reasons of convention and engineering.

    ANTsR function: `antsApplyTransformsToPoints`

    Arguments
    ---------
    dim: integer
         dimensionality of the transformation.

    points: data frame
          moving point set with n-points in rows of at least dim
          columns - we maintain extra information in additional
          columns. this should be a data frame with columns names x, y, z, t.

    transformlist : list of strings
        list of transforms generated by ants.registration where each transform is a filename.

    whichtoinvert : list of booleans (optional)
        Must be same length as transformlist.
        whichtoinvert[i] is True if transformlist[i] is a matrix,
        and the matrix should be inverted. If transformlist[i] is a
        warp field, whichtoinvert[i] must be False.
        If the transform list is a matrix followed by a warp field,
        whichtoinvert defaults to (True,False). Otherwise it defaults
        to [False]*len(transformlist)).

    verbose : boolean

    Returns
    -------
    data frame of transformed points

    Example
    -------
    >>> import ants
    >>> fixed = ants.image_read( ants.get_ants_data('r16') )
    >>> moving = ants.image_read( ants.get_ants_data('r27') )
    >>> reg = ants.registration( fixed, moving, 'Affine' )
    >>> d = {'x': [128, 127], 'y': [101, 111]}
    >>> pts = pd.DataFrame(data=d)
    >>> ptsw = ants.apply_transforms_to_points( 2, pts, reg['fwdtransforms'])
    """

    if not isinstance(transformlist, (tuple, list)) and (transformlist is not None):
        transformlist = [transformlist]

    args = [dim, points, transformlist, whichtoinvert]

    for tl_path in transformlist:
        if not os.path.exists(tl_path):
            raise Exception('Transform %s does not exist' % tl_path)

    mytx = []

    if whichtoinvert is None or (isinstance(whichtoinvert, (tuple,list)) and (sum([w is not None for w in whichtoinvert])==0)):
        if (len(transformlist) == 2) and ('.mat' in transformlist[0]) and ('.mat' not in transformlist[1]):
            whichtoinvert = (True, False)
        else:
            whichtoinvert = tuple([False]*len(transformlist))

    if len(whichtoinvert) != len(transformlist):
        raise ValueError('Transform list and inversion list must be the same length')

    for i in range(len(transformlist)):
        ismat = False
        if '.mat' in transformlist[i]:
            ismat = True
        if whichtoinvert[i] and (not ismat):
            raise ValueError('Cannot invert transform %i (%s) because it is not a matrix' % (i, transformlist[i]))
        if whichtoinvert[i]:
            mytx = mytx + ['-t', '[%s,1]' % (transformlist[i])]
        else:
            mytx = mytx + ['-t', transformlist[i]]
    if dim == 2:
        pointsSub = points[['x','y']]
    if dim == 3:
        pointsSub = points[['x','y','z']]
    if dim == 4:
        pointsSub = points[['x','y','z','t']]
    pointImage = ants.make_image( pointsSub.shape, pointsSub.values.flatten())
    pointsOut = pointImage.clone()
    args = ['-d', dim,
            '-i', pointImage,
            '-o', pointsOut ]
    args = args + mytx
    myargs = process_arguments(args)

    myverb = int(verbose)
    if verbose:
        print(myargs)

    processed_args = myargs + [ '-f', str(1), '--precision', str(0)]
    libfn = get_lib_fn('antsApplyTransformsToPoints')
    libfn(processed_args)
    mynp = pointsOut.numpy()
    pointsOutDF = points.copy()
    pointsOutDF['x'] = mynp[:,0]
    if dim >= 2:
        pointsOutDF['y'] = mynp[:,1]
    if dim >= 3:
        pointsOutDF['z'] = mynp[:,2]
    if dim >= 4:
        pointsOutDF['t'] = mynp[:,3]
    return pointsOutDF



__all__ = ["hessian_objectness"]

import ants
from ants.decorators import image_method
from ants.internal import get_lib_fn

@image_method
def hessian_objectness(image,
                       object_dimension=1,
                       is_bright_object=True,
                       sigma_min=0.1,
                       sigma_max=10,
                       number_of_sigma_steps=10,
                       use_sigma_logarithmic_spacing=True,
                       alpha=0.5,
                       beta=0.5,
                       gamma=5.0,
                       set_scale_objectness_measure=True):
    """

    Interface to ITK filter.  Based on the paper by Westin et al., "Geometrical 
    Diffusion Measures for MRI from Tensor Basis Analysis" and Luca Antiga's 
    Insight Journal paper http://hdl.handle.net/1926/576.

    Arguments
    ---------
    image : ANTsImage
        scalar image.

    object_dimension : unsigned int
        0: 'sphere', 1: 'line', or 2: 'plane'.

    is_bright_object : boolean
        Set 'true' for enhancing bright objects and 'false' for dark objects.

    sigma_min : float
        Define scale domain for feature extraction.

    sigma_max : float
        Define scale domain for feature extraction.

    number_of_sigma_steps : unsigned int
        Define number of samples for scale space.
        
    use_sigma_logarithmic_spacing : boolean
        Define sample spacing the for scale space.

    alpha: float 
        Hessian filter parameter.
          
    beta: float 
        Hessian filter parameter.
       
    gamma: float 
        Hessian filter parameter.

    set_scale_objectness_measure: boolean
        ...     

    Returns
    -------
    ANTsImage

    Example
    -------
    >>> import ants
    >>> image = ants.image_read(ants.get_ants_data('r16'))
    >>> hessian_object_image = ants.hessian_objectness(image)
    """

    image_dimension = image.dimension

    libfn = get_lib_fn('hessianObjectnessF%i' % image_dimension)
    hessian = libfn(image.pointer, object_dimension, is_bright_object, sigma_min,
                                 sigma_max, number_of_sigma_steps,
                                 use_sigma_logarithmic_spacing, alpha,
                                 beta, gamma, set_scale_objectness_measure)
    output_image = ants.from_pointer(hessian).clone('float')
    return output_image





__all__ = ['eig_seg',
           'initialize_eigenanatomy',
           'sparse_decom2']

import numpy as np
from scipy.stats import pearsonr
import pandas as pd
from pandas import DataFrame
import statsmodels.api as sm

import ants
from ants.internal import get_lib_fn, get_pointer_string

def sparse_decom2(inmatrix,
                    inmask=(None, None),
                    sparseness=(0.01, 0.01),
                    nvecs=3,
                    its=20,
                    cthresh=(0,0),
                    statdir=None,
                    perms=0,
                    uselong=0,
                    z=0,
                    smooth=0,
                    robust=0,
                    mycoption=0,
                    initialization_list=[],
                    initialization_list2=[],
                    ell1=10,
                    prior_weight=0,
                    verbose=False,
                    rejector=0,
                    max_based=False,
                    version=1):
    """
    Decomposes two matrices into paired sparse eigenevectors to
    maximize canonical correlation - aka Sparse CCA.
    Note: we do not scale the matrices internally. We leave
    scaling choices to the user.

    ANTsR function: `sparseDecom2`

    Arguments
    ---------
    inmatrix : 2-tuple of ndarrays
        input as inmatrix=(mat1,mat2). n by p input matrix and n by q input matrix , spatial variable lies along columns.

    inmask : 2-tuple of ANTsImage types (optional - one or both)
        optional pair of image masks

    sparseness : tuple
        a pair of float values e.g c(0.01,0.1) enforces an unsigned 99 percent and 90 percent sparse solution for each respective view

    nvecs : integer
        number of eigenvector pairs

    its : integer
        number of iterations, 10 or 20 usually sufficient

    cthresh : 2-tuple
        cluster threshold pair

    statdir : string (optional)
        temporary directory if you want to look at full output

    perms : integer
        number of permutations. settings permutations greater than 0 will estimate significance per vector empirically. For small datasets, these may be conservative. p-values depend on how one scales the input matrices.

    uselong : boolean
        enforce solutions of both views to be the same - requires matrices to be the same size

    z : float
        subject space (low-dimensional space) sparseness value

    smooth : float
        smooth the data (only available when mask is used)

    robust : boolean
        rank transform input matrices

    mycoption : integer
        enforce 1 - spatial orthogonality, 2 - low-dimensional orthogonality or 0 - both

    initialization_list : list
        initialization for first view

    initialization_list2 : list
        initialization for 2nd view

    ell1 : float
        gradient descent parameter, if negative then l0 otherwise use l1

    prior_weight : scalar
        Scalar value weight on prior between 0 (prior is weak) and 1 (prior is strong). Only engaged if initialization is used

    verbose : boolean
        activates verbose output to screen

    rejector : scalar
        rejects small correlation solutions

    max_based : boolean
        whether to choose max-based thresholding

    Returns
    -------
    dict w/ following key/value pairs:

        `projections` : ndarray
            X projections

        `projections2` : ndarray
            Y projections

        `eig1` : ndarray
            X components

        `eig2` : ndarray
            Y components

        `summary` : pd.DataFrame
            first column is canonical correlations,
            second column is p-values (these are `None` if perms > 0)

    Example
    -------
    >>> import numpy as np
    >>> import ants
    >>> mat = np.random.randn(20, 100)
    >>> mat2 = np.random.randn(20, 90)
    >>> mydecom = ants.sparse_decom2(inmatrix = (mat,mat2),
                                    sparseness=(0.1,0.3), nvecs=3,
                                    its=3, perms=0)

    """
    if inmatrix[0].shape[0] != inmatrix[1].shape[0]:
        raise ValueError('Matrices must have same number of rows (samples)')

    idim = 3

    if ants.is_image(inmask[0]):
        maskx = inmask[0].clone('float')
        idim = inmask[0].dimension
        hasmaskx = 1
    elif isinstance(inmask[0], np.ndarray):
        maskx = ants.from_numpy(inmask[0], pixeltype='float')
        idim = inmask[0].ndim
        hasmaskx = 1
    else:
        maskx = ants.make_image([1]*idim, pixeltype='float')
        hasmaskx = -1

    if ants.is_image(inmask[1]):
        masky = inmask[1].clone('float')
        idim = inmask[1].dimension
        hasmasky = 1
    elif isinstance(inmask[1], np.ndarray):
        masky = ants.from_numpy(inmask[1], pixeltype='float')
        idim = inmask[1].ndim
        hasmasky = 1
    else:
        masky = ants.make_image([1]*idim, pixeltype='float')
        hasmasky = -1

    inmask = [maskx, masky]

    if robust > 0:
        raise NotImplementedError('robust > 0 not currently implemented')
    else:
        input_matrices = inmatrix

    if idim == 2:
        if version == 1:
            sccancpp_fn = get_lib_fn('sccanCpp2D')
        elif version == 2:
            sccancpp_fn = get_lib_fn('sccanCpp2DV2')
            input_matrices = (input_matrices[0].tolist(), input_matrices[1].tolist())
    elif idim ==3:
        if version == 1:
            sccancpp_fn = get_lib_fn('sccanCpp3D')
        elif version == 2:
            sccancpp_fn = get_lib_fn('sccanCpp3DV2')
            input_matrices = (input_matrices[0].tolist(), input_matrices[1].tolist())

    outval = sccancpp_fn(input_matrices[0], input_matrices[1],
                        inmask[0].pointer, inmask[1].pointer,
                        hasmaskx, hasmasky,
                        sparseness[0], sparseness[1],
                        nvecs, its,
                        cthresh[0], cthresh[1],
                        z, smooth,
                        initialization_list, initialization_list2,
                        ell1, verbose,
                        prior_weight, mycoption, max_based)
    outval['eig1'] = np.array(outval['eig1'])
    outval['eig2'] = np.array(outval['eig2'])

    p1 = np.dot(input_matrices[0], outval['eig1'].T)
    p2 = np.dot(input_matrices[1], outval['eig2'].T)
    outcorrs = np.array([pearsonr(p1[:,i],p2[:,i])[0] for i in range(p1.shape[1])])
    if prior_weight < 1e-10:
        myord = np.argsort(np.abs(outcorrs))[::-1]
        outcorrs = outcorrs[myord]
        p1 = p1[:, myord]
        p2 = p2[:, myord]
        outval['eig1'] = outval['eig1'][myord,:]
        outval['eig2'] = outval['eig2'][myord,:]

    cca_summary = np.vstack((outcorrs,[None]*len(outcorrs))).T

    if perms > 0:
        cca_summary[:,1] = 0

        nsubs = input_matrices[0].shape[0]
        for permer in range(perms):
            m1 = input_matrices[0][np.random.permutation(nsubs),:]
            m2 = input_matrices[1][np.random.permutation(nsubs),:]
            outvalperm = sccancpp_fn(m1, m2,
                                inmask[0].pointer, inmask[1].pointer,
                                hasmaskx, hasmasky,
                                sparseness[0], sparseness[1],
                                nvecs, its,
                                cthresh[0], cthresh[1],
                                z, smooth,
                                initialization_list, initialization_list2,
                                ell1, verbose,
                                prior_weight, mycoption, max_based)
            outvalperm['eig1'] = np.array(outvalperm['eig1'])
            outvalperm['eig2'] = np.array(outvalperm['eig2'])
            p1perm = np.dot(m1, outvalperm['eig1'].T)
            p2perm = np.dot(m2, outvalperm['eig2'].T)
            outcorrsperm = np.array([pearsonr(p1perm[:,i],p2perm[:,i])[0] for i in range(p1perm.shape[1])])
            if prior_weight < 1e-10:
                myord = np.argsort(np.abs(outcorrsperm))[::-1]
                outcorrsperm = outcorrsperm[myord]
            counter = np.abs(cca_summary[:,0]) < np.abs(outcorrsperm)
            counter = counter.astype('int')
            cca_summary[:,1] = cca_summary[:,1] + counter

        cca_summary[:,1] = cca_summary[:,1] / float(perms)

    return {'projections': p1,
            'projections2': p2,
            'eig1': outval['eig1'].T,
            'eig2': outval['eig2'].T,
            'summary': pd.DataFrame(cca_summary,columns=['corrs','pvalues'])}


def initialize_eigenanatomy(initmat, mask=None, initlabels=None, nreps=1, smoothing=0):
    """
    InitializeEigenanatomy is a helper function to initialize sparseDecom
    and sparseDecom2. Can be used to estimate sparseness parameters per
    eigenvector. The user then only chooses nvecs and optional
    regularization parameters.

    Arguments
    ---------
    initmat : np.ndarray or ANTsImage
        input matrix where rows provide initial vector values.
        alternatively, this can be an antsImage which contains labeled regions.

    mask : ANTsImage
        mask if available

    initlabels : list/tuple of integers
        which labels in initmat to use as initial components

    nreps : integer
        nrepetitions to use

    smoothing : float
        if using an initial label image, optionally smooth each roi

    Returns
    -------
    dict w/ the following key/value pairs:
        `initlist` : list of ANTsImage types
            initialization list(s) for sparseDecom(2)

        `mask` : ANTsImage
            mask(s) for sparseDecom(2)

        `enames` : list of strings
            string names of components for sparseDecom(2)

    Example
    -------
    >>> import ants
    >>> import numpy as np
    >>> mat = np.random.randn(4,100).astype('float32')
    >>> init = ants.initialize_eigenanatomy(mat)
    """
    if ants.is_image(initmat):
        # create initmat from each of the unique labels
        if mask is not None:
            selectvec = mask > 0
        else:
            selectvec = initmat > 0
        initmatvec = initmat[selectvec]

        if initlabels is None:
            ulabs = np.sort(np.unique(initmatvec))
            ulabs = ulabs[ulabs > 0]
        else:
            ulabs = initlabels

        nvox = len(initmatvec)
        temp = np.zeros((len(ulabs), nvox))

        for x in range(len(ulabs)):
            timg = ants.threshold_image(initmat, ulabs[x]-1e-4, ulabs[x]+1e-4)
            if smoothing > 0:
                timg = ants.smooth_image(timg, smoothing)
            temp[x,:] = timg[selectvec]
        initmat = temp

    nclasses = initmat.shape[0]
    classlabels = ['init%i'%i for i in range(nclasses)]
    initlist = []
    if mask is None:
        maskmat = np.zeros(initmat.shape)
        maskmat[0,:] = 1
        mask = ants.from_numpy(maskmat.astype('float32'))

    eanatnames = ['A'] * (nclasses*nreps)
    ct = 0
    for i in range(nclasses):
        vecimg = mask.clone('float')
        initf = initmat[i,:]
        vecimg[mask==1] = initf
        for nr in range(nreps):
            initlist.append(vecimg)
            eanatnames[ct+nr-1] = str(classlabels[i])
            ct = ct + 1

    return {'initlist': initlist, 'mask': mask, 'enames': eanatnames}



def eig_seg(mask, img_list, apply_segmentation_to_images=False, cthresh=0, smooth=1):
    """
    Segment a mask into regions based on the max value in an image list.
    At a given voxel the segmentation label will contain the index to the image
    that has the largest value. If the 3rd image has the greatest value,
    the segmentation label will be 3 at that voxel.

    Arguments
    ---------
    mask : ANTsImage
        D-dimensional mask > 0 defining segmentation region.

    img_list : collection of ANTsImage or np.ndarray
        images to use

    apply_segmentation_to_images : boolean
        determines if original image list is modified by the segmentation.

    cthresh : integer
        throw away isolated clusters smaller than this value

    smooth : float
        smooth the input data first by this value

    Returns
    -------
    ANTsImage

    Example
    -------
    >>> import ants
    >>> mylist = [ants.image_read(ants.get_ants_data('r16')),
                  ants.image_read(ants.get_ants_data('r27')),
                  ants.image_read(ants.get_ants_data('r85'))]
    >>> myseg = ants.eig_seg(ants.get_mask(mylist[0]), mylist)
    """
    maskvox = mask > 0
    maskseg = mask.clone()
    maskseg[maskvox] = 0
    if isinstance(img_list, np.ndarray):
        mydata = img_list
    elif isinstance(img_list, (tuple, list)):
        mydata = ants.image_list_to_matrix(img_list, mask)

    if (smooth > 0):
        for i in range(mydata.shape[0]):
            temp_img = ants.make_image(mask, mydata[i,:], pixeltype='float')
            temp_img = ants.smooth_image(temp_img, smooth, sigma_in_physical_coordinates=True)
            mydata[i,:] = temp_img[mask >= 0.5]

    segids = np.argmax(np.abs(mydata), axis=0)+1
    segmax = np.max(np.abs(mydata), axis=0)
    maskseg[maskvox] = (segids * (segmax > 1e-09))

    if cthresh > 0:
        for kk in range(int(maskseg.max())):
            timg = ants.threshold_image(maskseg, kk, kk)
            timg = ants.label_clusters(timg, cthresh)
            timg = ants.threshold_image(timg, 1, 1e15) * float(kk)
            maskseg[maskseg == kk] = timg[maskseg == kk]

    if (apply_segmentation_to_images) and (not isinstance(img_list, np.ndarray)):
        for i in range(len(img_list)):
            img = img_list[i]
            img[maskseg != float(i)] = 0
            img_list[i] = img

    return maskseg



"""
Functions for plotting ants images
"""


__all__ = [
    "plot_ortho_stack"
]

import fnmatch
import math
import os
import warnings

from matplotlib import gridspec
import matplotlib.pyplot as plt
import matplotlib.patheffects as path_effects
import matplotlib.lines as mlines
import matplotlib.patches as patches
import matplotlib.mlab as mlab
import matplotlib.animation as animation
from mpl_toolkits.axes_grid1.inset_locator import inset_axes


import numpy as np
import ants




def plot_ortho_stack(
    images,
    overlays=None,
    reorient=True,
    # xyz arguments
    xyz=None,
    xyz_lines=False,
    xyz_color="red",
    xyz_alpha=0.6,
    xyz_linewidth=2,
    xyz_pad=5,
    # base image arguments
    cmap="Greys_r",
    alpha=1,
    # overlay arguments
    overlay_cmap="jet",
    overlay_alpha=0.9,
    # background arguments
    black_bg=True,
    bg_thresh_quant=0.01,
    bg_val_quant=0.99,
    # scale/crop/domain arguments
    crop=False,
    scale=False,
    domain_image_map=None,
    # title arguments
    title=None,
    titlefontsize=24,
    title_dx=0,
    title_dy=0,
    # 4th panel text arguemnts
    text=None,
    textfontsize=24,
    textfontcolor="white",
    text_dx=0,
    text_dy=0,
    # save & size arguments
    filename=None,
    dpi=500,
    figsize=1.0,
    colpad=0,
    rowpad=0,
    transpose=False,
    transparent=True,
    orient_labels=True,
):
    """
    Create a stack of orthographic plots with optional overlays.

    Use mask_image and/or threshold_image to preprocess images to be be
    overlaid and display the overlays in a given range. See the wiki examples.

    Example
    -------
    >>> import ants
    >>> mni = ants.image_read(ants.get_data('mni'))
    >>> ch2 = ants.image_read(ants.get_data('ch2'))
    >>> ants.plot_ortho_stack([mni,mni,mni])
    """

    def mirror_matrix(x):
        return x[::-1, :]

    def rotate270_matrix(x):
        return mirror_matrix(x.T)

    def reorient_slice(x, axis):
        return rotate270_matrix(x)

    # need this hack because of a weird NaN warning from matplotlib with overlays
    warnings.simplefilter("ignore")

    n_images = len(images)

    # handle `image` argument
    for i in range(n_images):
        if isinstance(images[i], str):
            images[i] = ants.image_read(images[i])
        if not ants.is_image(images[i]):
            raise ValueError("image argument must be an ANTsImage")
        if images[i].dimension != 3:
            raise ValueError("Input image must have 3 dimensions!")

    if overlays is None:
        overlays = [None] * n_images
    # handle `overlay` argument
    for i in range(n_images):
        if overlays[i] is not None:
            if isinstance(overlays[i], str):
                overlays[i] = ants.image_read(overlays[i])
            if not ants.is_image(overlays[i]):
                raise ValueError("overlay argument must be an ANTsImage")
            if overlays[i].components > 1:
                raise ValueError("overlays[i] cannot have more than one voxel component")
            if overlays[i].dimension != 3:
                raise ValueError("Overlay image must have 3 dimensions!")

            if not ants.image_physical_space_consistency(images[i], overlays[i]):
                overlays[i] = ants.resample_image_to_target(
                    overlays[i], images[i], interp_type="linear"
                )

    for i in range(1, n_images):
        if not ants.image_physical_space_consistency(images[0], images[i]):
            images[i] = ants.resample_image_to_target(
                images[0], images[i], interp_type="linear"
            )

    # reorient images
    if reorient != False:
        if reorient == True:
            reorient = "RPI"

        for i in range(n_images):
            images[i] = images[i].reorient_image2(reorient)

        if overlays[i] is not None:
            overlays[i] = overlays[i].reorient_image2(reorient)

    # handle `slices` argument
    if xyz is None:
        xyz = [int(s / 2) for s in images[0].shape]
    for i in range(3):
        if xyz[i] is None:
            xyz[i] = int(images[0].shape[i] / 2)

    # resample image if spacing is very unbalanced
    spacing = [s for i, s in enumerate(images[0].spacing)]
    if (max(spacing) / min(spacing)) > 3.0:
        new_spacing = (1, 1, 1)
        for i in range(n_images):
            images[i] = images[i].resample_image(tuple(new_spacing))
        if overlays[i] is not None:
            overlays[i] = overlays[i].resample_image(tuple(new_spacing))
        xyz = [
            int(sl * (sold / snew)) for sl, sold, snew in zip(xyz, spacing, new_spacing)
        ]

    # potentially crop image
    if crop:
        for i in range(n_images):
            plotmask = images[i].get_mask(cleanup=0)
            if plotmask.max() == 0:
                plotmask += 1
            images[i] = images[i].crop_image(plotmask)
            if overlays[i] is not None:
                overlays[i] = overlays[i].crop_image(plotmask)

    # pad images
    for i in range(n_images):
        if i == 0:
            images[i], lowpad, uppad = images[i].pad_image(return_padvals=True)
        else:
            images[i] = images[i].pad_image()
        if overlays[i] is not None:
            overlays[i] = overlays[i].pad_image()
    xyz = [v + l for v, l in zip(xyz, lowpad)]

    # handle `domain_image_map` argument
    if domain_image_map is not None:
        if ants.is_image(domain_image_map):
            tx = ants.new_ants_transform(
                precision="float", transform_type="AffineTransform", dimension=3
            )
            for i in range(n_images):
                images[i] = ants.apply_ants_transform_to_image(
                    tx, images[i], domain_image_map
                )

                if overlays[i] is not None:
                    overlays[i] = ants.apply_ants_transform_to_image(
                        tx, overlays[i], domain_image_map, interpolation="linear"
                    )
        else:
            raise Exception('The domain_image_map must be an ants image.')

    # potentially find dynamic range
    if scale == True:
        vmins = []
        vmaxs = []
        for i in range(n_images):
            vmin, vmax = images[i].quantile((0.05, 0.95))
            vmins.append(vmin)
            vmaxs.append(vmax)
    elif isinstance(scale, (list, tuple)):
        if len(scale) != 2:
            raise ValueError(
                "scale argument must be boolean or list/tuple with two values"
            )
        vmins = []
        vmaxs = []
        for i in range(n_images):
            vmin, vmax = images[i].quantile(scale)
            vmins.append(vmin)
            vmaxs.append(vmax)
    else:
        vmin = None
        vmax = None

    if not transpose:
        nrow = n_images
        ncol = 3
    else:
        nrow = 3
        ncol = n_images

    fig = plt.figure(figsize=((ncol + 1) * 2.5 * figsize, (nrow + 1) * 2.5 * figsize))
    if title is not None:
        basey = 0.93
        basex = 0.5
        fig.suptitle(
            title, fontsize=titlefontsize, color=textfontcolor, x=basex + title_dx, y=basey + title_dy
        )

    if (colpad > 0) and (rowpad > 0):
        bothgridpad = max(colpad, rowpad)
        colpad = 0
        rowpad = 0
    else:
        bothgridpad = 0.0

    gs = gridspec.GridSpec(
        nrow,
        ncol,
        wspace=bothgridpad,
        hspace=0.0,
        top=1.0 - 0.5 / (nrow + 1),
        bottom=0.5 / (nrow + 1) + colpad,
        left=0.5 / (ncol + 1) + rowpad,
        right=1 - 0.5 / (ncol + 1),
    )

    # pad image to have isotropic array dimensions
    vminols=[]
    vmaxols=[]
    for i in range(n_images):
        images[i] = images[i].numpy()
        if overlays[i] is not None:
            vminols.append( overlays[i].min() )
            vmaxols.append( overlays[i].max() )
            overlays[i] = overlays[i].numpy()
            if overlays[i].dtype not in ["uint8", "uint32"]:
                overlays[i][np.abs(overlays[i]) == 0] = np.nan

    ####################
    ####################
    for i in range(n_images):
        yz_slice = reorient_slice(images[i][xyz[0], :, :], 0)
        if not transpose:
            ax = plt.subplot(gs[i, 0])
        else:
            ax = plt.subplot(gs[0, i])
        ax.imshow(yz_slice, cmap=cmap, vmin=vmin, vmax=vmax)
        if overlays[i] is not None:
            yz_overlay = reorient_slice(overlays[i][xyz[0], :, :], 0)
            ax.imshow(yz_overlay, alpha=overlay_alpha, cmap=overlay_cmap,
                vmin=vminols[i], vmax=vmaxols[i])
        if xyz_lines:
            # add lines
            l = mlines.Line2D(
                [xyz[1], xyz[1]],
                [xyz_pad, yz_slice.shape[0] - xyz_pad],
                color=xyz_color,
                alpha=xyz_alpha,
                linewidth=xyz_linewidth,
            )
            ax.add_line(l)
            l = mlines.Line2D(
                [xyz_pad, yz_slice.shape[1] - xyz_pad],
                [yz_slice.shape[1] - xyz[2], yz_slice.shape[1] - xyz[2]],
                color=xyz_color,
                alpha=xyz_alpha,
                linewidth=xyz_linewidth,
            )
            ax.add_line(l)
        if orient_labels:
            ax.text(
                0.5,
                0.98,
                "S",
                horizontalalignment="center",
                verticalalignment="top",
                fontsize=20 * figsize,
                color=textfontcolor,
                transform=ax.transAxes,
            )
            ax.text(
                0.5,
                0.02,
                "I",
                horizontalalignment="center",
                verticalalignment="bottom",
                fontsize=20 * figsize,
                color=textfontcolor,
                transform=ax.transAxes,
            )
            ax.text(
                0.98,
                0.5,
                "A",
                horizontalalignment="right",
                verticalalignment="center",
                fontsize=20 * figsize,
                color=textfontcolor,
                transform=ax.transAxes,
            )
            ax.text(
                0.02,
                0.5,
                "P",
                horizontalalignment="left",
                verticalalignment="center",
                fontsize=20 * figsize,
                color=textfontcolor,
                transform=ax.transAxes,
            )
        ax.axis("off")
        ####################
        ####################

        xz_slice = reorient_slice(images[i][:, xyz[1], :], 1)
        if not transpose:
            ax = plt.subplot(gs[i, 1])
        else:
            ax = plt.subplot(gs[1, i])
        ax.imshow(xz_slice, cmap=cmap, vmin=vmin, vmax=vmax)
        if overlays[i] is not None:
            xz_overlay = reorient_slice(overlays[i][:, xyz[1], :], 1)
            ax.imshow(xz_overlay, alpha=overlay_alpha, cmap=overlay_cmap,
                vmin=vminols[i], vmax=vmaxols[i])
        if xyz_lines:
            # add lines
            l = mlines.Line2D(
                [xz_slice.shape[0] - xyz[0], xz_slice.shape[0] - xyz[0]],
                [xyz_pad, xz_slice.shape[0] - xyz_pad],
                color=xyz_color,
                alpha=xyz_alpha,
                linewidth=xyz_linewidth,
            )
            ax.add_line(l)
            l = mlines.Line2D(
                [xyz_pad, xz_slice.shape[1] - xyz_pad],
                [xz_slice.shape[1] - xyz[2], xz_slice.shape[1] - xyz[2]],
                color=xyz_color,
                alpha=xyz_alpha,
                linewidth=xyz_linewidth,
            )
            ax.add_line(l)
        if orient_labels:
            ax.text(
                0.5,
                0.98,
                "I",
                horizontalalignment="center",
                verticalalignment="top",
                fontsize=20 * figsize,
                color=textfontcolor,
                transform=ax.transAxes,
            )
            ax.text(
                0.5,
                0.02,
                "S",
                horizontalalignment="center",
                verticalalignment="bottom",
                fontsize=20 * figsize,
                color=textfontcolor,
                transform=ax.transAxes,
            )
            ax.text(
                0.98,
                0.5,
                "L",
                horizontalalignment="right",
                verticalalignment="center",
                fontsize=20 * figsize,
                color=textfontcolor,
                transform=ax.transAxes,
            )
            ax.text(
                0.02,
                0.5,
                "R",
                horizontalalignment="left",
                verticalalignment="center",
                fontsize=20 * figsize,
                color=textfontcolor,
                transform=ax.transAxes,
            )
        ax.axis("off")

        ####################
        ####################
        xy_slice = reorient_slice(images[i][:, :, xyz[2]], 2)
        if not transpose:
            ax = plt.subplot(gs[i, 2])
        else:
            ax = plt.subplot(gs[2, i])
        ax.imshow(xy_slice, cmap=cmap, vmin=vmin, vmax=vmax)
        if overlays[i] is not None:
            xy_overlay = reorient_slice(overlays[i][:, :, xyz[2]], 2)
            ax.imshow(xy_overlay, alpha=overlay_alpha, cmap=overlay_cmap,
                vmin=vminols[i], vmax=vmaxols[i])
        if xyz_lines:
            # add lines
            l = mlines.Line2D(
                [xy_slice.shape[0] - xyz[0], xy_slice.shape[0] - xyz[0]],
                [xyz_pad, xy_slice.shape[0] - xyz_pad],
                color=xyz_color,
                alpha=xyz_alpha,
                linewidth=xyz_linewidth,
            )
            ax.add_line(l)
            l = mlines.Line2D(
                [xyz_pad, xy_slice.shape[1] - xyz_pad],
                [xy_slice.shape[1] - xyz[1], xy_slice.shape[1] - xyz[1]],
                color=xyz_color,
                alpha=xyz_alpha,
                linewidth=xyz_linewidth,
            )
            ax.add_line(l)
        if orient_labels:
            ax.text(
                0.5,
                0.98,
                "A",
                horizontalalignment="center",
                verticalalignment="top",
                fontsize=20 * figsize,
                color=textfontcolor,
                transform=ax.transAxes,
            )
            ax.text(
                0.5,
                0.02,
                "P",
                horizontalalignment="center",
                verticalalignment="bottom",
                fontsize=20 * figsize,
                color=textfontcolor,
                transform=ax.transAxes,
            )
            ax.text(
                0.98,
                0.5,
                "L",
                horizontalalignment="right",
                verticalalignment="center",
                fontsize=20 * figsize,
                color=textfontcolor,
                transform=ax.transAxes,
            )
            ax.text(
                0.02,
                0.5,
                "R",
                horizontalalignment="left",
                verticalalignment="center",
                fontsize=20 * figsize,
                color=textfontcolor,
                transform=ax.transAxes,
            )
        ax.axis("off")

    ####################
    ####################

    if filename is not None:
        plt.savefig(filename, dpi=dpi, transparent=transparent)
        plt.close(fig)
    else:
        plt.show()

    # turn warnings back to default
    warnings.simplefilter("default")



import os
import unittest
from common import run_tests

import ants

import numpy as np
import numpy.testing as nptest


class Test_iMath(unittest.TestCase):
    
    
    def setUp(self):
        pass
    
    def tearDown(self):
        pass
    
    def test_functions(self):
        image = ants.image_read(ants.get_data('r16'))
        ants.iMath_canny(image, sigma=2, lower=0, upper=1)
        ants.iMath_fill_holes(image, hole_type=2)
        ants.iMath_GC(image, radius=1)
        ants.iMath_GD(image, radius=1)
        ants.iMath_GE(image, radius=1)
        ants.iMath_GO(image, radius=1)
        ants.iMath_get_largest_component(image, min_size=50)
        ants.iMath_grad(image, sigma=0.5, normalize=False)
        ants.iMath_histogram_equalization(image, alpha=0.5, beta=0.5)
        ants.iMath_laplacian(image, sigma=0.5, normalize=False)
        ants.iMath_MC(image, radius=1, value=1, shape=1, parametric=False, lines=3, thickness=1, include_center=False)
        ants.iMath_MD(image, radius=1, value=1, shape=1, parametric=False, lines=3, thickness=1, include_center=False)
        ants.iMath_ME(image, radius=1, value=1, shape=1, parametric=False, lines=3, thickness=1, include_center=False)
        ants.iMath_MO(image, radius=1, value=1, shape=1, parametric=False, lines=3, thickness=1, include_center=False)
        ants.iMath_maurer_distance(image, foreground=1)
        ants.iMath_normalize(image)
        ants.iMath_pad(image, padding=2)
        ants.iMath_perona_malik(image, conductance=0.25, n_iterations=1)
        ants.iMath_sharpen(image)
        ants.iMath_truncate_intensity(image, lower_q=0.05, upper_q=0.95, n_bins=64)
    
    
class Test_slice_image(unittest.TestCase):
    """
    Test ants.ANTsImage class
    """
    def setUp(self):
        pass

    def tearDown(self):
        pass

    def test_slice_image_2d(self):
        """
        Test that resampling an image doesnt cause the resampled
        image to have NaNs - previously caused by resampling an
        image of type DOUBLE
        """
        img = ants.image_read(ants.get_ants_data('r16'))
        
        img2 = ants.slice_image(img, 0, 100)
        self.assertTrue(isinstance(img2, np.ndarray))
        
        img2 = ants.slice_image(img, 1, 100)
        self.assertTrue(isinstance(img2, np.ndarray))
        
        with self.assertRaises(Exception):
            img2 = ants.slice_image(img, 2, 100)
            
        with self.assertRaises(Exception):
            img2 = ants.slice_image(img, -3, 100)
        
        with self.assertRaises(Exception):
            img2 = ants.slice_image(img, 4, 100)
            
        img2 = ants.slice_image(img, -1, 100)
        img3 = ants.slice_image(img, 1, 100)
        self.assertTrue(np.allclose(img2, img3))

    def test_slice_image_2d_vector(self):
        img0 = ants.image_read(ants.get_ants_data('r16'))
        img = ants.merge_channels([img0,img0,img0])

        img2 = ants.slice_image(img, 0, 100)
        self.assertTrue(isinstance(img2, np.ndarray))
        
        img2 = ants.slice_image(img, 1, 100)
        self.assertTrue(isinstance(img2, np.ndarray))
        
        with self.assertRaises(Exception):
            img2 = ants.slice_image(img, 2, 100)
            
        with self.assertRaises(Exception):
            img2 = ants.slice_image(img, -3, 100)
        
        with self.assertRaises(Exception):
            img2 = ants.slice_image(img, 4, 100)
            
        img2 = ants.slice_image(img, -1, 100)
        img3 = ants.slice_image(img, 1, 100)
        self.assertTrue(np.allclose(img2, img3))

    def test_slice_image_3d(self):
        """
        Test that resampling an image doesnt cause the resampled
        image to have NaNs - previously caused by resampling an
        image of type DOUBLE
        """
        img = ants.image_read(ants.get_ants_data('mni'))
        
        img2 = ants.slice_image(img, 0, 100)
        self.assertEqual(img2.dimension, 2)
        
        img2 = ants.slice_image(img, 1, 100)
        self.assertEqual(img2.dimension, 2)
        
        img2 = ants.slice_image(img, 2, 100)
        self.assertEqual(img2.dimension, 2)
        
        img2 = ants.slice_image(img.clone('unsigned int'), 2, 100)
        self.assertEqual(img2.dimension, 2)
        
        with self.assertRaises(Exception):
            img2 = ants.slice_image(img, 3, 100)
            
        with self.assertRaises(Exception):
            img2 = ants.slice_image(img, 2, 100, collapse_strategy=23)
            
        img2 = ants.slice_image(img, -1, 100)
        img3 = ants.slice_image(img, 2, 100)
        self.assertTrue(ants.allclose(img2, img3))

    def test_slice_image_3d_vector(self):
        """
        Test that resampling an image doesnt cause the resampled
        image to have NaNs - previously caused by resampling an
        image of type DOUBLE
        """
        img0 = ants.image_read(ants.get_ants_data('mni'))
        img = ants.merge_channels([img0,img0,img0])
        
        img2 = ants.slice_image(img, 0, 100)
        self.assertEqual(img2.dimension, 2)
        
        img2 = ants.slice_image(img, 1, 100)
        self.assertEqual(img2.dimension, 2)
        
        img2 = ants.slice_image(img, 2, 100)
        self.assertEqual(img2.dimension, 2)
        
        img2 = ants.slice_image(img.clone('unsigned int'), 2, 100)
        self.assertEqual(img2.dimension, 2)
        
        with self.assertRaises(Exception):
            img2 = ants.slice_image(img, 3, 100)
            
        with self.assertRaises(Exception):
            img2 = ants.slice_image(img, 2, 100, collapse_strategy=23)
            
        img2 = ants.slice_image(img, -1, 100)
        img3 = ants.slice_image(img, 2, 100)
        self.assertTrue(ants.allclose(img2, img3))


if __name__ == '__main__':
    run_tests()



__all__ = ["label_overlap_measures"]

import pandas as pd


from ants.internal import get_lib_fn
from ants.decorators import image_method

@image_method
def label_overlap_measures(source_image, target_image):
    """
    Get overlap measures from two label images (e.g., Dice)

    ANTsR function: `labelOverlapMeasures`

    Arguments
    ---------
    source image : ANTsImage
        Source image

    target_image : ANTsImage
        Target image

    Returns
    -------
    data frame with measures for each label and all labels combined

    Example
    -------
    >>> import ants
    >>> r16 = ants.image_read( ants.get_ants_data('r16') )
    >>> r64 = ants.image_read( ants.get_ants_data('r64') )
    >>> s16 = ants.kmeans_segmentation( r16, 3 )['segmentation']
    >>> s64 = ants.kmeans_segmentation( r64, 3 )['segmentation']
    >>> stats = ants.label_overlap_measures(s16, s64)
    """
    source_image_int = source_image.clone("unsigned int")
    target_image_int = target_image.clone("unsigned int")

    libfn = get_lib_fn("labelOverlapMeasures%iD" % source_image_int.dimension)
    df = libfn(source_image_int.pointer, target_image_int.pointer)
    df = pd.DataFrame(df)
    # Set Label column to object type as it can contain strings
    df['Label'] = df['Label'].astype(object)
    df.loc[0, 'Label'] = 'All'
    return df



"""
Various data augmentation transforms for ANTsImage types

List of Transformations:
======================
- CastIntensity
- BlurIntensity
- NormalizeIntensity
- RescaleIntensity
- ShiftScaleIntensity
- SigmoidIntensity
======================
- FlipImage
- TranslateImage

TODO
----
- RotateImage
- ShearImage
- ScaleImage
- DeformImage
- PadImage
- HistogramEqualizeIntensity
- TruncateIntensity
- SharpenIntensity
- MorpholigicalIntensity
    - MD
    - ME
    - MO
    - MC
    - GD
    - GE
    - GO
    - GC
"""
__all__ = ['CastIntensity',
           'BlurIntensity',
           'LocallyBlurIntensity',
           'NormalizeIntensity',
           'RescaleIntensity',
           'ShiftScaleIntensity',
           'SigmoidIntensity',
           'FlipImage',
           'ScaleImage',
           'TranslateImage',
           'MultiResolutionImage']

from ... import utils
from ...core import ants_image as iio


class MultiResolutionImage(object):
    """
    Generate a set of images at multiple resolutions from an original image
    """
    def __init__(self, levels=4, keep_shape=False):
        self.levels = levels
        self.keep_shape = keep_shape

    def transform(self, X, y=None):
        """
        Generate a set of multi-resolution ANTsImage types

        Arguments
        ---------
        X : ANTsImage
            image to transform

        y : ANTsImage (optional)
            another image to transform

        Example
        -------
        >>> import ants
        >>> multires = ants.contrib.MultiResolutionImage(levels=4)
        >>> img = ants.image_read(ants.get_data('r16'))
        >>> imgs = multires.transform(img)
        """
        insuffix = X._libsuffix
        multires_fn = utils.get_lib_fn('multiResolutionAntsImage%s' % (insuffix))
        casted_ptrs = multires_fn(X.pointer, self.levels)
        
        imgs = []
        for casted_ptr in casted_ptrs:
            img = iio.ANTsImage(pixeltype=X.pixeltype, dimension=X.dimension,
                                components=X.components, pointer=casted_ptr)
            if self.keep_shape:
                img = img.resample_image_to_target(X)
            imgs.append(img)

        return imgs


## Intensity Transforms ##

class CastIntensity(object):
    """
    Cast the pixeltype of an ANTsImage to a given type. 
    This code uses the C++ ITK library directly, so it is fast.
    
    NOTE: This offers a ~2.5x speedup over using img.clone(pixeltype):
    
    Timings vs Cloning
    ------------------
    >>> import ants
    >>> import time
    >>> caster = ants.contrib.CastIntensity('float')
    >>> img = ants.image_read(ants.get_data('mni')).clone('unsigned int')
    >>> s = time.time()
    >>> for i in range(1000):
    ...     img_float = caster.transform(img)
    >>> e = time.time()
    >>> print(e - s) # 9.6s
    >>> s = time.time()
    >>> for i in range(1000):
    ...     img_float = img.clone('float')
    >>> e = time.time()
    >>> print(e - s) # 25.3s
    """
    def __init__(self, pixeltype):
        """
        Initialize a CastIntensity transform
    
        Arguments
        ---------
        pixeltype : string
            pixeltype to which images will be casted

        Example
        -------
        >>> import ants
        >>> caster = ants.contrib.CastIntensity('float')
        """
        self.pixeltype = pixeltype

    def transform(self, X, y=None):
        """
        Transform an image by casting its type

        Arguments
        ---------
        X : ANTsImage
            image to cast

        y : ANTsImage (optional)
            another image to cast.

        Example
        -------
        >>> import ants
        >>> caster = ants.contrib.CastIntensity('float')
        >>> img2d = ants.image_read(ants.get_data('r16')).clone('unsigned int')
        >>> img2d_float = caster.transform(img2d)
        >>> print(img2d.pixeltype, '- ', img2d_float.pixeltype)
        >>> img3d = ants.image_read(ants.get_data('mni')).clone('unsigned int')
        >>> img3d_float = caster.transform(img3d)
        >>> print(img3d.pixeltype, ' - ' , img3d_float.pixeltype)
        """
        insuffix = X._libsuffix
        outsuffix = '%s%i' % (utils.short_ptype(self.pixeltype), X.dimension)
        cast_fn = utils.get_lib_fn('castAntsImage%s%s' % (insuffix, outsuffix))
        casted_ptr = cast_fn(X.pointer)
        return iio.ANTsImage(pixeltype=self.pixeltype, dimension=X.dimension,
                             components=X.components, pointer=casted_ptr)


class BlurIntensity(object):
    """
    Transform for blurring the intensity of an ANTsImage
    using a Gaussian Filter
    """
    def __init__(self, sigma, width):
        """
        Initialize a BlurIntensity transform

        Arguments
        ---------
        sigma : float
            variance of gaussian kernel intensity
            increasing this value increasing the amount
            of blur

        width : int
            width of gaussian kernel shape
            increasing this value increase the number of
            neighboring voxels which are used for blurring

        Example
        -------
        >>> import ants
        >>> blur = ants.contrib.BlurIntensity(2,3)
        """
        self.sigma = sigma
        self.width = width

    def transform(self, X, y=None):
        """
        Blur an image by applying a gaussian filter.

        Arguments
        ---------
        X : ANTsImage
            image to transform

        y : ANTsImage (optional)
            another image to transform.

        Example
        -------
        >>> import ants
        >>> blur = ants.contrib.BlurIntensity(2,3)
        >>> img2d = ants.image_read(ants.get_data('r16'))
        >>> img2d_b = blur.transform(img2d)
        >>> ants.plot(img2d)
        >>> ants.plot(img2d_b)
        >>> img3d = ants.image_read(ants.get_data('mni'))
        >>> img3d_b = blur.transform(img3d)
        >>> ants.plot(img3d)
        >>> ants.plot(img3d_b)
        """
        if X.pixeltype != 'float':
            raise ValueError('image.pixeltype must be float ... use TypeCast transform or clone to float')

        insuffix = X._libsuffix
        cast_fn = utils.get_lib_fn('blurAntsImage%s' % (insuffix))
        casted_ptr = cast_fn(X.pointer, self.sigma, self.width)
        return iio.ANTsImage(pixeltype=X.pixeltype, dimension=X.dimension,
                             components=X.components, pointer=casted_ptr,
                             origin=X.origin)


class LocallyBlurIntensity(object):
    """
    Blur an ANTsImage locally using a gradient anisotropic
    diffusion filter, thereby preserving the sharpeness of edges as best 
    as possible.
    """
    def __init__(self, conductance=1, iters=5):
        self.conductance = conductance
        self.iters = iters

    def transform(self, X, y=None):
        """
        Locally blur an image by applying a gradient anisotropic diffusion filter.

        Arguments
        ---------
        X : ANTsImage
            image to transform

        y : ANTsImage (optional)
            another image to transform.

        Example
        -------
        >>> import ants
        >>> blur = ants.contrib.LocallyBlurIntensity(1,5)
        >>> img2d = ants.image_read(ants.get_data('r16'))
        >>> img2d_b = blur.transform(img2d)
        >>> ants.plot(img2d)
        >>> ants.plot(img2d_b)
        >>> img3d = ants.image_read(ants.get_data('mni'))
        >>> img3d_b = blur.transform(img3d)
        >>> ants.plot(img3d)
        >>> ants.plot(img3d_b)
        """
        #if X.pixeltype != 'float':
        #    raise ValueError('image.pixeltype must be float ... use TypeCast transform or clone to float')
        insuffix = X._libsuffix
        cast_fn = utils.get_lib_fn('locallyBlurAntsImage%s' % (insuffix))
        casted_ptr = cast_fn(X.pointer, self.iters, self.conductance)
        return iio.ANTsImage(pixeltype=X.pixeltype, dimension=X.dimension,
                             components=X.components, pointer=casted_ptr)


class NormalizeIntensity(object):
    """
    Normalize the intensity values of an ANTsImage to have
    zero mean and unit variance

    NOTE: this transform is more-or-less the same in speed
    as an equivalent numpy+scikit-learn solution.

    Timing vs Numpy+Scikit-Learn
    ----------------------------
    >>> import ants
    >>> import numpy as np
    >>> from sklearn.preprocessing import StandardScaler
    >>> import time
    >>> img = ants.image_read(ants.get_data('mni'))
    >>> arr = img.numpy().reshape(1,-1)
    >>> normalizer = ants.contrib.NormalizeIntensity()
    >>> normalizer2 = StandardScaler()
    >>> s = time.time()
    >>> for i in range(100):
    ...     img_scaled = normalizer.transform(img)
    >>> e = time.time()
    >>> print(e - s) # 3.3s
    >>> s = time.time()
    >>> for i in range(100):
    ...     arr_scaled = normalizer2.fit_transform(arr)
    >>> e = time.time()
    >>> print(e - s) # 3.5s
    """
    def __init__(self):
        """
        Initialize a NormalizeIntensity transform
        """
        pass

    def transform(self, X, y=None):
        """
        Transform an image by normalizing its intensity values to 
        have zero mean and unit variance.

        Arguments
        ---------
        X : ANTsImage
            image to transform

        y : ANTsImage (optional)
            another image to transform.

        Example
        -------
        >>> import ants
        >>> normalizer = ants.contrib.NormalizeIntensity()
        >>> img2d = ants.image_read(ants.get_data('r16'))
        >>> img2d_r = normalizer.transform(img2d)
        >>> print(img2d.mean(), ',', img2d.std(), ' -> ', img2d_r.mean(), ',', img2d_r.std())
        >>> img3d = ants.image_read(ants.get_data('mni'))
        >>> img3d_r = normalizer.transform(img3d)
        >>> print(img3d.mean(), ',' , img3d.std(), ',', ' -> ', img3d_r.mean(), ',' , img3d_r.std())  
        """
        if X.pixeltype != 'float':
            raise ValueError('image.pixeltype must be float ... use TypeCast transform or clone to float')

        insuffix = X._libsuffix
        cast_fn = utils.get_lib_fn('normalizeAntsImage%s' % (insuffix))
        casted_ptr = cast_fn(X.pointer)
        return iio.ANTsImage(pixeltype=X.pixeltype, dimension=X.dimension,
                             components=X.components, pointer=casted_ptr)


class RescaleIntensity(object):
    """
    Rescale the pixeltype of an ANTsImage linearly to be between a given
    minimum and maximum value. 
    This code uses the C++ ITK library directly, so it is fast.
    
    NOTE: this offered a ~5x speedup over using built-in arithmetic operations in ANTs.
    It is also more-or-less the same in speed as an equivalent numpy+scikit-learn
    solution.

    Timing vs Built-in Operations
    -----------------------------
    >>> import ants
    >>> import time
    >>> rescaler = ants.contrib.RescaleIntensity(0,1)
    >>> img = ants.image_read(ants.get_data('mni'))
    >>> s = time.time()
    >>> for i in range(100):
    ...     img_float = rescaler.transform(img)
    >>> e = time.time()
    >>> print(e - s) # 2.8s
    >>> s = time.time()
    >>> for i in range(100):
    ...     maxval = img.max()
    ...     img_float = (img - maxval) / (maxval - img.min())
    >>> e = time.time()
    >>> print(e - s) # 13.9s

    Timing vs Numpy+Scikit-Learn
    ----------------------------
    >>> import ants
    >>> import numpy as np
    >>> from sklearn.preprocessing import MinMaxScaler
    >>> import time
    >>> img = ants.image_read(ants.get_data('mni'))
    >>> arr = img.numpy().reshape(1,-1)
    >>> rescaler = ants.contrib.RescaleIntensity(-1,1)
    >>> rescaler2 = MinMaxScaler((-1,1)).fit(arr)
    >>> s = time.time()
    >>> for i in range(100):
    ...     img_scaled = rescaler.transform(img)
    >>> e = time.time()
    >>> print(e - s) # 2.8s
    >>> s = time.time()
    >>> for i in range(100):
    ...     arr_scaled = rescaler2.transform(arr)
    >>> e = time.time()
    >>> print(e - s) # 3s
    """

    def __init__(self, min_val, max_val):
        """
        Initialize a RescaleIntensity transform.

        Arguments
        ---------
        min_val : float
            minimum value to which image(s) will be rescaled

        max_val : float
            maximum value to which image(s) will be rescaled
        
        Example
        -------
        >>> import ants
        >>> rescaler = ants.contrib.RescaleIntensity(0,1)
        """
        self.min_val = min_val
        self.max_val = max_val

    def transform(self, X, y=None):
        """
        Transform an image by linearly rescaling its intensity to
        be between a minimum and maximum value

        Arguments
        ---------
        X : ANTsImage
            image to transform

        y : ANTsImage (optional)
            another image to transform.

        Example
        -------
        >>> import ants
        >>> rescaler = ants.contrib.RescaleIntensity(0,1)
        >>> img2d = ants.image_read(ants.get_data('r16'))
        >>> img2d_r = rescaler.transform(img2d)
        >>> print(img2d.min(), ',', img2d.max(), ' -> ', img2d_r.min(), ',', img2d_r.max())
        >>> img3d = ants.image_read(ants.get_data('mni'))
        >>> img3d_r = rescaler.transform(img3d)
        >>> print(img3d.min(), ',' , img3d.max(), ' -> ', img3d_r.min(), ',' , img3d_r.max())
        """
        if X.pixeltype != 'float':
            raise ValueError('image.pixeltype must be float ... use TypeCast transform or clone to float')

        insuffix = X._libsuffix
        cast_fn = utils.get_lib_fn('rescaleAntsImage%s' % (insuffix))
        casted_ptr = cast_fn(X.pointer, self.min_val, self.max_val)
        return iio.ANTsImage(pixeltype=X.pixeltype, dimension=X.dimension,
                             components=X.components, pointer=casted_ptr)


class ShiftScaleIntensity(object):
    """
    Shift and scale the intensity of an ANTsImage
    """
    def __init__(self, shift, scale):
        """
        Initialize a ShiftScaleIntensity transform

        Arguments
        ---------
        shift : float
            shift all of the intensity values by the given amount through addition.
            For example, if the minimum image value is 0.0 and the shift
            is 10.0, then the new minimum value (before scaling) will be 10.0

        scale : float
            scale all the intensity values by the given amount through multiplication.
            For example, if the min/max image values are 10/20 and the scale
            is 2.0, then then new min/max values will be 20/40

        Example
        -------
        >>> import ants
        >>> shiftscaler = ants.contrib.ShiftScaleIntensity(shift=10, scale=2)
        """
        self.shift = shift
        self.scale = scale

    def transform(self, X, y=None):
        """
        Transform an image by shifting and scaling its intensity values.

        Arguments
        ---------
        X : ANTsImage
            image to transform

        y : ANTsImage (optional)
            another image to transform.

        Example
        -------
        >>> import ants
        >>> shiftscaler = ants.contrib.ShiftScaleIntensity(10,2.)
        >>> img2d = ants.image_read(ants.get_data('r16'))
        >>> img2d_r = shiftscaler.transform(img2d)
        >>> print(img2d.min(), ',', img2d.max(), ' -> ', img2d_r.min(), ',', img2d_r.max())
        >>> img3d = ants.image_read(ants.get_data('mni'))
        >>> img3d_r = shiftscaler.transform(img3d)
        >>> print(img3d.min(), ',' , img3d.max(), ',', ' -> ', img3d_r.min(), ',' , img3d_r.max())  
        """
        if X.pixeltype != 'float':
            raise ValueError('image.pixeltype must be float ... use TypeCast transform or clone to float')

        insuffix = X._libsuffix
        cast_fn = utils.get_lib_fn('shiftScaleAntsImage%s' % (insuffix))
        casted_ptr = cast_fn(X.pointer, self.scale, self.shift)
        return iio.ANTsImage(pixeltype=X.pixeltype, dimension=X.dimension,
                             components=X.components, pointer=casted_ptr)


class SigmoidIntensity(object):
    """
    Transform an image using a sigmoid function
    """
    def __init__(self, min_val, max_val, alpha, beta):
        """
        Initialize a SigmoidIntensity transform

        Arguments
        ---------
        min_val : float
            minimum value

        max_val : float
            maximum value

        alpha : float
            alpha value for sigmoid

        beta : flaot
            beta value for sigmoid

        Example
        -------
        >>> import ants
        >>> sigscaler = ants.contrib.SigmoidIntensity(0,1,1,1)
        """
        self.min_val = min_val
        self.max_val = max_val
        self.alpha = alpha
        self.beta = beta

    def transform(self, X, y=None):
        """
        Transform an image by applying a sigmoid function.

        Arguments
        ---------
        X : ANTsImage
            image to transform

        y : ANTsImage (optional)
            another image to transform.

        Example
        -------
        >>> import ants
        >>> sigscaler = ants.contrib.SigmoidIntensity(0,1,1,1)
        >>> img2d = ants.image_read(ants.get_data('r16'))
        >>> img2d_r = sigscaler.transform(img2d)
        >>> img3d = ants.image_read(ants.get_data('mni'))
        >>> img3d_r = sigscaler.transform(img3d)
        """
        if X.pixeltype != 'float':
            raise ValueError('image.pixeltype must be float ... use TypeCast transform or clone to float')

        insuffix = X._libsuffix
        cast_fn = utils.get_lib_fn('sigmoidAntsImage%s' % (insuffix))
        casted_ptr = cast_fn(X.pointer, self.min_val, self.max_val, self.alpha, self.beta)
        return iio.ANTsImage(pixeltype=X.pixeltype, dimension=X.dimension,
                             components=X.components, pointer=casted_ptr)


## Physical Transforms ##

class FlipImage(object):
    """
    Transform an image by flipping two axes. 
    """
    def __init__(self, axis1, axis2):
        """
        Initialize a SigmoidIntensity transform

        Arguments
        ---------
        axis1 : int
            axis to flip

        axis2 : int
            other axis to flip

        Example
        -------
        >>> import ants
        >>> flipper = ants.contrib.FlipImage(0,1)
        """
        self.axis1 = axis1
        self.axis2 = axis2

    def transform(self, X, y=None):
        """
        Transform an image by applying a sigmoid function.

        Arguments
        ---------
        X : ANTsImage
            image to transform

        y : ANTsImage (optional)
            another image to transform.

        Example
        -------
        >>> import ants
        >>> flipper = ants.contrib.FlipImage(0,1)
        >>> img2d = ants.image_read(ants.get_data('r16'))
        >>> img2d_r = flipper.transform(img2d)
        >>> ants.plot(img2d)
        >>> ants.plot(img2d_r)
        >>> flipper2 = ants.contrib.FlipImage(1,0)
        >>> img2d = ants.image_read(ants.get_data('r16'))
        >>> img2d_r = flipper2.transform(img2d)
        >>> ants.plot(img2d)
        >>> ants.plot(img2d_r)
        """
        if X.pixeltype != 'float':
            raise ValueError('image.pixeltype must be float ... use TypeCast transform or clone to float')

        insuffix = X._libsuffix
        cast_fn = utils.get_lib_fn('flipAntsImage%s' % (insuffix))
        casted_ptr = cast_fn(X.pointer, self.axis1, self.axis2)
        return iio.ANTsImage(pixeltype=X.pixeltype, dimension=X.dimension,
                             components=X.components, pointer=casted_ptr,
                             origin=X.origin)


class TranslateImage(object):
    """
    Translate an image in physical space. This function calls 
    highly optimized ITK/C++ code.
    """
    def __init__(self, translation, reference=None, interp='linear'):
        """
        Initialize a TranslateImage transform

        Arguments
        ---------
        translation : list, tuple, or numpy.ndarray
            absolute pixel transformation in each axis

        reference : ANTsImage (optional)
            image which provides the reference physical space in which
            to perform the transform

        interp : string
            type of interpolation to use
            options: linear, nearest

        Example
        -------
        >>> import ants
        >>> translater = ants.contrib.TranslateImage((10,10), interp='linear')
        """
        if interp not in {'linear', 'nearest'}:
            raise ValueError('interp must be one of {linear, nearest}')

        self.translation = list(translation)
        self.reference = reference
        self.interp = interp

    def transform(self, X, y=None):
        """
        Example
        -------
        >>> import ants
        >>> translater = ants.contrib.TranslateImage((40,0))
        >>> img2d = ants.image_read(ants.get_data('r16'))
        >>> img2d_r = translater.transform(img2d)
        >>> ants.plot(img2d, img2d_r)
        >>> translater = ants.contrib.TranslateImage((40,0,0))
        >>> img3d = ants.image_read(ants.get_data('mni'))
        >>> img3d_r = translater.transform(img3d)
        >>> ants.plot(img3d, img3d_r, axis=2)
        """
        if X.pixeltype != 'float':
            raise ValueError('image.pixeltype must be float ... use TypeCast transform or clone to float')

        if len(self.translation) != X.dimension:
            raise ValueError('must give a translation value for each image dimension')

        if self.reference is None:
            reference = X
        else:
            reference = self.reference

        insuffix = X._libsuffix
        cast_fn = utils.get_lib_fn('translateAntsImage%s_%s' % (insuffix, self.interp))
        casted_ptr = cast_fn(X.pointer, reference.pointer, self.translation)
        return iio.ANTsImage(pixeltype=X.pixeltype, dimension=X.dimension,
                             components=X.components, pointer=casted_ptr)


class ScaleImage(object):
    """
    Scale an image in physical space. This function calls 
    highly optimized ITK/C++ code.
    """
    def __init__(self, scale, reference=None, interp='linear'):
        """
        Initialize a TranslateImage transform

        Arguments
        ---------
        scale : list, tuple, or numpy.ndarray
            relative scaling along each axis

        reference : ANTsImage (optional)
            image which provides the reference physical space in which
            to perform the transform

        interp : string
            type of interpolation to use
            options: linear, nearest

        Example
        -------
        >>> import ants
        >>> translater = ants.contrib.TranslateImage((10,10), interp='linear')
        """
        if interp not in {'linear', 'nearest'}:
            raise ValueError('interp must be one of {linear, nearest}')

        self.scale = list(scale)
        self.reference = reference
        self.interp = interp

    def transform(self, X, y=None):
        """
        Example
        -------
        >>> import ants
        >>> scaler = ants.contrib.ScaleImage((1.2,1.2))
        >>> img2d = ants.image_read(ants.get_data('r16'))
        >>> img2d_r = scaler.transform(img2d)
        >>> ants.plot(img2d, img2d_r)
        >>> scaler = ants.contrib.ScaleImage((1.2,1.2,1.2))
        >>> img3d = ants.image_read(ants.get_data('mni'))
        >>> img3d_r = scaler.transform(img3d)
        >>> ants.plot(img3d, img3d_r)
        """
        if X.pixeltype != 'float':
            raise ValueError('image.pixeltype must be float ... use TypeCast transform or clone to float')

        if len(self.scale) != X.dimension:
            raise ValueError('must give a scale value for each image dimension')

        if self.reference is None:
            reference = X
        else:
            reference = self.reference

        insuffix = X._libsuffix
        cast_fn = utils.get_lib_fn('scaleAntsImage%s_%s' % (insuffix, self.interp))
        casted_ptr = cast_fn(X.pointer, reference.pointer, self.scale)
        return iio.ANTsImage(pixeltype=X.pixeltype, dimension=X.dimension,
                             components=X.components, pointer=casted_ptr)







__all__ = ['get_orientation',
           'reorient_image2',
           'get_possible_orientations',
           'get_center_of_mass']

import numpy as np
from tempfile import mktemp

import ants
from ants.decorators import image_method
from ants.internal import get_lib_fn



_possible_orientations = ['RIP','LIP',  'RSP',  'LSP',  'RIA',  'LIA',
'RSA',  'LSA',  'IRP',  'ILP',  'SRP',  'SLP',  'IRA',  'ILA',  'SRA',
'SLA',  'RPI',  'LPI',  'RAI',  'LAI',  'RPS',  'LPS',  'RAS',  'LAS',
'PRI',  'PLI',  'ARI',  'ALI',  'PRS',  'PLS',  'ARS',  'ALS',  'IPR',
'SPR',  'IAR',  'SAR',  'IPL',  'SPL',  'IAL',  'SAL',  'PIR',  'PSR',
'AIR',  'ASR',  'PIL',  'PSL',  'AIL',  'ASL']


def get_possible_orientations():
    return _possible_orientations

@image_method
def get_orientation(image):
    direction = image.direction

    orientation = []
    for i in range(3):
        row = direction[:,i]
        idx = np.where(np.abs(row)==np.max(np.abs(row)))[0][0]

        if idx == 0:
            if row[idx] < 0:
                orientation.append('L')
            else:
                orientation.append('R')
        elif idx == 1:
            if row[idx] < 0:
                orientation.append('P')
            else:
                orientation.append('A')
        elif idx == 2:
            if row[idx] < 0:
                orientation.append('S')
            else:
                orientation.append('I')
    return ''.join(orientation)

@image_method
def reorient_image2(image, orientation='RAS'):
     """
     Reorient an image.
     Example
     -------
     >>> import ants
     >>> mni = ants.image_read(ants.get_data('mni'))
     >>> mni2 = mni.reorient_image2()
     """
     if image.has_components:
         return ants.merge_channels([img.reorient_image2(orientation) for img in ants.split_channels(image)],
                                    channels_first=image.channels_first)
     
     if image.dimension != 3:
         raise ValueError('image must have 3 dimensions')

     inpixeltype = image.pixeltype
     ndim = image.dimension
     if image.pixeltype != 'float':
         image = image.clone('float')

     libfn = get_lib_fn('reorientImage2')
     itkimage = libfn(image.pointer, orientation)

     new_img = ants.from_pointer(itkimage)
     if inpixeltype != 'float':
         new_img = new_img.clone(inpixeltype)
     return new_img
     
@image_method
def get_center_of_mass(image):
    """
    Compute an image center of mass in physical space which is defined
    as the mean of the intensity weighted voxel coordinate system.

    ANTsR function: `getCenterOfMass`

    Arguments
    ---------
    image : ANTsImage
        image from which center of mass will be computed

    Returns
    -------
    scalar

    Example
    -------
    >>> fi = ants.image_read( ants.get_ants_data("r16"))
    >>> com1 = ants.get_center_of_mass( fi )
    >>> fi = ants.image_read( ants.get_ants_data("r64"))
    >>> com2 = ants.get_center_of_mass( fi )
    """
    if image.pixeltype != 'float':
        image = image.clone('float')

    libfn = get_lib_fn('centerOfMass')
    com = libfn(image.pointer)

    return tuple(com)




__all__ = ['get_neighborhood_in_mask',
           'get_neighborhood_at_voxel']

import numpy as np

import ants

from ants.internal import get_lib_fn
from ants.decorators import image_method

@image_method
def get_neighborhood_in_mask(image, mask, radius, physical_coordinates=False,
                            boundary_condition=None, spatial_info=False, get_gradient=False):
    """
    Get neighborhoods for voxels within mask.
    
    This converts a scalar image to a matrix with rows that contain neighbors 
    around a center voxel
    
    ANTsR function: `getNeighborhoodInMask`

    Arguments
    ---------
    image : ANTsImage
        image to get values from
    
    mask : ANTsImage
        image indicating which voxels to examine. Each voxel > 0 will be used as the 
        center of a neighborhood
    
    radius : tuple/list
        array of values for neighborhood radius (in voxels)
    
    physical_coordinates : boolean
        whether voxel indices and offsets should be in voxel or physical coordinates
    
    boundary_condition : string (optional)
        how to handle voxels in a neighborhood, but not in the mask.
            None : fill values with `NaN`
            `image` : use image value, even if not in mask
            `mean` : use mean of all non-NaN values for that neighborhood
    
    spatial_info : boolean
        whether voxel locations and neighborhood offsets should be returned along with pixel values.
    
    get_gradient : boolean
        whether a matrix of gradients (at the center voxel) should be returned in 
        addition to the value matrix (WIP)

    Returns
    -------
    if spatial_info is False:
        if get_gradient is False:
            ndarray
                an array of pixel values where the number of rows is the size of the 
                neighborhood and there is a column for each voxel

        else if get_gradient is True:
            dictionary w/ following key-value pairs:
                values : ndarray
                    array of pixel values where the number of rows is the size of the 
                    neighborhood and there is a column for each voxel.

                gradients : ndarray
                    array providing the gradients at the center voxel of each 
                    neighborhood
        
    else if spatial_info is True:
        dictionary w/ following key-value pairs:
            values : ndarray
                array of pixel values where the number of rows is the size of the 
                neighborhood and there is a column for each voxel.

            indices : ndarray
                array provinding the center coordinates for each neighborhood

            offsets : ndarray
                array providing the offsets from center for each voxel in a neighborhood

    Example
    -------
    >>> import ants
    >>> r16 = ants.image_read(ants.get_ants_data('r16'))
    >>> mask = ants.get_mask(r16)
    >>> mat = ants.get_neighborhood_in_mask(r16, mask, radius=(2,2))
    """
    if not ants.is_image(image):
        raise ValueError('image must be ANTsImage type')
    if not ants.is_image(mask):
        raise ValueError('mask must be ANTsImage type')
    if isinstance(radius, (int, float)):
        radius = [radius]*image.dimension
    if (not isinstance(radius, (tuple,list))) or (len(radius) != image.dimension):
        raise ValueError('radius must be tuple or list with length == image.dimension')

    boundary = 0
    if boundary_condition == 'image':
        boundary = 1
    elif boundary_condition == 'mean':
        boundary = 2
    
    libfn = get_lib_fn('getNeighborhoodMatrix%s' % image._libsuffix)
    retvals = libfn(image.pointer, 
                    mask.pointer, 
                    list(radius),
                    int(physical_coordinates), 
                    int(boundary),
                    int(spatial_info),
                    int(get_gradient))

    if not spatial_info:
        if get_gradient:
            retvals['values'] = np.asarray(retvals['values'])
            retvals['gradients'] = np.asarray(retvals['gradients'])
        else:
            retvals = np.asarray(retvals['matrix'])
    else:
        retvals['values'] = np.asarray(retvals['values'])
        retvals['indices'] = np.asarray(retvals['indices'])
        retvals['offsets'] = np.asarray(retvals['offsets'])

    return retvals

@image_method
def get_neighborhood_at_voxel(image, center, kernel, physical_coordinates=False):
    """
    Get a hypercube neighborhood at a voxel. Get the values in a local 
    neighborhood of an image.
    
    ANTsR function: `getNeighborhoodAtVoxel`

    Arguments
    ---------
    image : ANTsImage
        image to get values from.
    
    center : tuple/list
        indices for neighborhood center
    
    kernel : tuple/list
        either a collection of values for neighborhood radius (in voxels) or 
        a binary collection of the same dimension as the image, specifying the shape of the neighborhood to extract
    
    physical_coordinates : boolean
        whether voxel indices and offsets should be in voxel 
        or physical coordinates

    Returns
    -------
    dictionary w/ following key-value pairs:
        values : ndarray
            array of neighborhood values at the voxel

        indices : ndarray
            matrix providing the coordinates for each value

    Example
    -------
    >>> import ants
    >>> img = ants.image_read(ants.get_ants_data('r16'))
    >>> center = (2,2)
    >>> radius = (3,3)
    >>> retval = ants.get_neighborhood_at_voxel(img, center, radius)
    """
    if not ants.is_image(image):
        raise ValueError('image must be ANTsImage type')

    if (not isinstance(center, (tuple,list))) or (len(center) != image.dimension):
        raise ValueError('center must be tuple or list with length == image.dimension')

    if (not isinstance(kernel, (tuple,list))) or (len(kernel) != image.dimension):
        raise ValueError('kernel must be tuple or list with length == image.dimension')

    radius = [int((k-1)/2) for k in kernel]

    libfn = get_lib_fn('getNeighborhood%s' % image._libsuffix)
    retvals = libfn(image.pointer, 
                    list(center), 
                    list(kernel), 
                    list(radius), 
                    int(physical_coordinates))
    for k in retvals.keys():
        retvals[k] = np.asarray(retvals[k])
    return retvals





__all__ = ["build_template"]

import numpy as np
import os
import shutil
from tempfile import mktemp

import ants

def build_template(
    initial_template=None,
    image_list=None,
    iterations=3,
    gradient_step=0.2,
    blending_weight=0.75,
    weights=None,
    useNoRigid=True,
    output_dir=None,
    **kwargs
):
    """
    Estimate an optimal template from an input image_list

    ANTsR function: N/A

    Arguments
    ---------
    initial_template : ANTsImage
        initialization for the template building

    image_list : ANTsImages
        images from which to estimate template

    iterations : integer
        number of template building iterations

    gradient_step : scalar
        for shape update gradient

    blending_weight : scalar
        weight for image blending

    weights : vector
        weight for each input image

    useNoRigid : boolean
        equivalent of -y in the script. Template update
        step will not use the rigid component if this is True.

    output_dir : path
        directory name where intermediate transforms are written

    kwargs : keyword args
        extra arguments passed to ants registration

    Returns
    -------
    ANTsImage

    Example
    -------
    >>> import ants
    >>> image = ants.image_read( ants.get_ants_data('r16') )
    >>> image2 = ants.image_read( ants.get_ants_data('r27') )
    >>> image3 = ants.image_read( ants.get_ants_data('r85') )
    >>> timage = ants.build_template( image_list = ( image, image2, image3 ) ).resample_image( (45,45))
    >>> timagew = ants.build_template( image_list = ( image, image2, image3 ), weights = (5,1,1) )
    """
    work_dir = mktemp() if output_dir is None else output_dir

    def make_outprefix(k: int):
        os.makedirs(os.path.join(work_dir, f"img{k:04d}"), exist_ok=True)
        return os.path.join(work_dir, f"img{k:04d}", "out")

    if "type_of_transform" not in kwargs:
        type_of_transform = "SyN"
    else:
        type_of_transform = kwargs.pop("type_of_transform")

    if weights is None:
        weights = np.repeat(1.0 / len(image_list), len(image_list))
    weights = [x / sum(weights) for x in weights]
    if initial_template is None:
        initial_template = image_list[0] * 0
        for i in range(len(image_list)):
            temp = image_list[i] * weights[i]
            temp = ants.resample_image_to_target(temp, initial_template)
            initial_template = initial_template + temp

    xavg = initial_template.clone()
    for i in range(iterations):
        affinelist = []
        for k in range(len(image_list)):
            w1 = ants.registration(
                xavg, image_list[k], type_of_transform=type_of_transform, outprefix=make_outprefix(k), **kwargs
            )
            L = len(w1["fwdtransforms"])
            # affine is the last one
            affinelist.append(w1["fwdtransforms"][L-1])

            if k == 0:
                if L == 2:
                    wavg = ants.image_read(w1["fwdtransforms"][0]) * weights[k]
                xavgNew = w1["warpedmovout"] * weights[k]
            else:
                if L == 2:
                    wavg = wavg + ants.image_read(w1["fwdtransforms"][0]) * weights[k]
                xavgNew = xavgNew + w1["warpedmovout"] * weights[k]

        if useNoRigid:
            avgaffine = ants.average_affine_transform_no_rigid(affinelist)
        else:
            avgaffine = ants.average_affine_transform(affinelist)
        afffn = os.path.join(work_dir, "avgAffine.mat")
        ants.write_transform(avgaffine, afffn)

        if L == 2:
            print(wavg.abs().mean())
            wscl = (-1.0) * gradient_step
            wavg = wavg * wscl
            # apply affine to the nonlinear?
            # need to save the average
            wavgA = ants.apply_transforms(fixed=xavgNew, moving=wavg, imagetype=1, transformlist=afffn, whichtoinvert=[1])
            wavgfn = os.path.join(work_dir, "avgWarp.nii.gz")
            ants.image_write(wavgA, wavgfn)
            xavg = ants.apply_transforms(fixed=xavgNew, moving=xavgNew, transformlist=[wavgfn, afffn], whichtoinvert=[0, 1])
        else:
            xavg = ants.apply_transforms(fixed=xavgNew, moving=xavgNew, transformlist=[afffn], whichtoinvert=[1])
            
        if blending_weight is not None:
            xavg = xavg * blending_weight + ants.iMath(xavg, "Sharpen") * (
                1.0 - blending_weight
            )

    if output_dir is None:
        shutil.rmtree(work_dir)
    return xavg




__version__ = '0.6.1'

from .core import *
from .label import *
from .learn import *
from .math import *
from .ops import *
from .plotting import *
from .registration import *
from .segmentation import *
from .utils import *
from .contrib import *
from .deeplearn import *






__all__ = ['get_mask']

import ants
from ants.decorators import image_method

@image_method
def get_mask(image, low_thresh=None, high_thresh=None, cleanup=2):
    """
    Get a binary mask image from the given image after thresholding

    ANTsR function: `getMask`

    Arguments
    ---------
    image : ANTsImage
        image from which mask will be computed. Can be an antsImage of 2, 3 or 4 dimensions.

    low_thresh : scalar (optional)
        An inclusive lower threshold for voxels to be included in the mask.
        If not given, defaults to image mean.

    high_thresh : scalar (optional)
        An inclusive upper threshold for voxels to be included in the mask.
        If not given, defaults to image max

    cleanup : integer
        If > 0, morphological operations will be applied to clean up the mask by eroding away small or weakly-connected areas, and closing holes.
        If cleanup is >0, the following steps are applied
            1. Erosion with radius 2 voxels
            2. Retain largest component
            3. Dilation with radius 1 voxel
            4. Morphological closing

    Returns
    -------
    ANTsImage

    Example
    -------
    >>> import ants
    >>> image = ants.image_read( ants.get_ants_data('r16') )
    >>> mask = ants.get_mask(image)
    """
    cleanup = int(cleanup)
    if ants.is_image(image):
        if image.pixeltype != 'float':
            image = image.clone('float')

    if low_thresh is None:
        low_thresh = image.mean()
    if high_thresh is None:
        high_thresh = image.max()

    mask_image = ants.threshold_image(image, low_thresh, high_thresh)
    if cleanup > 0:
        mask_image = ants.iMath(mask_image, 'ME', cleanup)
        mask_image = ants.iMath(mask_image, 'GetLargestComponent')
        mask_image = ants.iMath(mask_image, 'MD', cleanup)
        mask_image = ants.iMath(mask_image, 'FillHoles').threshold_image( 1, 2 )
        while ((mask_image.min() == mask_image.max()) and (cleanup > 0)):
            cleanup = cleanup - 1
            mask_image = ants.threshold_image(image, low_thresh, high_thresh)
            if cleanup > 0:
                mask_image = ants.iMath(mask_image, 'ME', cleanup)
                mask_image = ants.iMath(mask_image, 'MD', cleanup)
                mask_image = ants.iMath(mask_image, 'FillHoles').threshold_image( 1, 2 )

            #if cleanup == 0:
            #    clustlab = ants.label_clusters(mask_image, 1)
            #    mask_image = ants.threshold_image(clustlab, 1, 1)

    return mask_image




__all__ = ['functional_lung_segmentation']

import ants

def functional_lung_segmentation(image,
                                 mask=None,
                                 number_of_iterations=2,
                                 number_of_atropos_iterations=5,
                                 mrf_parameters="[0.7,2x2x2]",
                                 number_of_clusters = 6,
                                 cluster_centers = None,
                                 bias_correction = "n4",
                                 verbose=True):

    """
    Ventilation-based segmentation of hyperpolarized gas lung MRI.

    Lung segmentation into classes based on ventilation as described in
    this paper:

         https://pubmed.ncbi.nlm.nih.gov/21837781/

    Arguments
    ---------
    image : ANTs image
        Input proton-weighted MRI.

    mask : ANTs image
        Mask image designating the region to segment.  0/1 = background/foreground.

    number_of_iterations : integer
        Number of Atropos <--> bias correction iterations (outer loop).

    number_of_atropos_iterations : integer
        Number of Atropos iterations (inner loop).  If number_of_atropos_iterations = 0,
        this is equivalent to K-means with no MRF priors.

    mrf_parameters : string
        Parameters for MRF in Atropos.

    number_of_clusters : integer
        Number of tissue classes.

    cluster_centers: array or tuple
        Initialization centers for k-means.

    bias_correction: string
        Apply "n3", "n4", or no bias correction (default = "n4").

    verbose : boolean
        Print progress to the screen.

    Returns
    -------
    Dictionary with segmentation image, probability images, and
    processed image.

    Example
    -------
    >>> import ants
    >>> image = ants.image_read(ants.get_data("mni")).resample_image((4,4,4))
    >>> mask = image.get_mask()
    >>> seg = ants.functional_lung_segmentation(image, mask, verbose=True,
                                                number_of_iterations=1,
                                                number_of_clusters=2,
                                                number_of_atropos_iterations=1)
    """

    if image.dimension != 3:
        raise ValueError("Function only works for 3-D images.")

    if mask is None:
        raise ValueError("Mask is missing.")

    if number_of_iterations < 1:
        raise ValueError("number_of_iterations must be >= 1.")

    def generate_pure_tissue_n4_weight_mask(probability_images):
        number_of_probability_images = len(probability_images)

        pure_tissue_mask = probability_images[0] * 0
        for i in range(number_of_probability_images):
            negation_image = probability_images[0] * 0 + 1
            for j in range(number_of_probability_images):
                if i != j:
                    negation_image = negation_image * (probability_images[j] * -1.0 + 1.0)
                pure_tissue_mask = pure_tissue_mask + negation_image * probability_images[i]
        return(pure_tissue_mask)

    weight_mask = None

    # This is a multiplicative factor for both the image
    # and the cluster centers which shouldn't effect the
    # user.  Otherwise, we'll get a singular covariance
    # complaint from Atropos.
    image_scale_factor = 1000.0

    number_of_atropos_n4_iterations = number_of_iterations
    for i in range(number_of_atropos_n4_iterations):
        if verbose == True:
            print("Outer iteration: ", i, " out of ", number_of_atropos_n4_iterations)

        preprocessed_image = ants.image_clone(image)

        quantiles = (preprocessed_image.quantile(0.0), preprocessed_image.quantile(0.995))
        preprocessed_image[preprocessed_image < quantiles[0]] = quantiles[0]
        preprocessed_image[preprocessed_image > quantiles[1]] = quantiles[1]

        if verbose == True:
            print("Outer: bias correction.")

        if bias_correction.lower() == "n4":
            preprocessed_image = ants.n4_bias_field_correction(preprocessed_image,
                mask=mask, shrink_factor=2, convergence={'iters': [50, 50, 50, 50], 'tol': 0.0000000001},
                spline_param=200, return_bias_field=False, weight_mask=weight_mask, verbose=verbose)
        elif bias_correction.lower == "n3":
            preprocessed_image = ants.n3_bias_field_correction(preprocessed_image, downsample_factor=2)

        preprocessed_image = ((preprocessed_image - preprocessed_image.min())
                              /(preprocessed_image.max() - preprocessed_image.min()))
        preprocessed_image *= image_scale_factor

        if verbose == True:
            print("Outer: Atropos segmentation.")

        atropos_initialization = "Kmeans[" + str(number_of_clusters) + "]"
        if cluster_centers is not None:
            if len(cluster_centers) != number_of_clusters:
                raise ValueError("number_of_clusters should match the vector size of the cluster_centers.")
            else:
                scaled_cluster_centers = image_scale_factor * cluster_centers
                cluster_centers_string = 'x'.join(str(s) for s in set(scaled_cluster_centers))
                atropos_initialization = "Kmeans[" + str(number_of_clusters) + "," + cluster_centers_string + "]"

        posterior_formulation = "Socrates[0]"
        if i > 0:
            atropos_initialization = atropos_output['probabilityimages']
            posterior_formulation = "Socrates[1]"

        iterations = "[" + str(number_of_atropos_iterations) + ",0]"
        atropos_verbose = 0
        if verbose == True:
            atropos_verbose = 1
        atropos_output = ants.atropos(preprocessed_image, x=mask, i=atropos_initialization,
            m=mrf_parameters, c=iterations, priorweight=0.0, v=atropos_verbose, p=posterior_formulation)

        weight_mask = generate_pure_tissue_n4_weight_mask(atropos_output['probabilityimages'][1:number_of_clusters])

    masked_segmentation_image = atropos_output['segmentation'] * mask
    masked_probability_images = list()
    for i in range(len(atropos_output['probabilityimages'])):
        masked_probability_images.append(atropos_output['probabilityimages'][i] * mask)

    return_dict = {'segmentation_image' : masked_segmentation_image,
                   'probability_images' : masked_probability_images,
                   'processed_image' : (preprocessed_image / image_scale_factor)}
    return(return_dict)



from functools import wraps
from ants.core.ants_image import ANTsImage

def image_method(func):
    @wraps(func) 
    def wrapper(self, *args, **kwargs): 
        return func(self, *args, **kwargs)
    setattr(ANTsImage, func.__name__, wrapper)
    return func




__all__ = ['weingarten_image_curvature']

import numpy as np

import ants
from ants.decorators import image_method
from ants.internal import get_lib_fn

@image_method
def weingarten_image_curvature(image, sigma=1.0, opt='mean'):
    """
    Uses the weingarten map to estimate image mean or gaussian curvature

    ANTsR function: `weingartenImageCurvature`
    
    Arguments
    ---------
    image : ANTsImage
        image from which curvature is calculated
    
    sigma : scalar
        smoothing parameter
    
    opt : string
        mean by default, otherwise `gaussian` or `characterize`
    
    Returns
    -------
    ANTsImage

    Example
    -------
    >>> import ants
    >>> image = ants.image_read(ants.get_ants_data('mni')).resample_image((3,3,3))
    >>> imagecurv = ants.weingarten_image_curvature(image)
    """
    if image.dimension not in {2,3}:
        raise ValueError('image must be 2D or 3D')

    if image.dimension == 2:
        d = image.shape
        temp = np.zeros(list(d)+[10])
        for k in range(1,7):
            voxvals = image[:d[0],:d[1]].numpy()
            temp[:d[0],:d[1],k] = voxvals
        temp = ants.from_numpy(temp)
        myspc = image.spacing
        myspc = list(myspc) + [min(myspc)]
        temp.set_spacing(myspc)
        temp = temp.clone('float')
    else:
        temp = image.clone('float')

    optnum = 0
    if opt == 'gaussian': 
        optnum = 6
    if opt == 'characterize':
        optnum = 5

    libfn = get_lib_fn('weingartenImageCurvature')
    mykout = libfn(temp.pointer, sigma, optnum)
    mykout = ants.from_pointer(mykout)
    if image.dimension == 3:
        return mykout
    elif image.dimension == 2:
        subarr = ants.from_numpy(mykout.numpy()[:,:,4])
        return ants.copy_image_info(image, subarr)





__all__ = ['image_similarity']

import ants
from ants.decorators import image_method

@image_method
def image_similarity(fixed_image, moving_image, metric_type='MeanSquares',
                    fixed_mask=None, moving_mask=None,
                    sampling_strategy='regular', sampling_percentage=1.):
    """
    Measure similarity between two images.
    NOTE: Similarity is actually returned as distance (i.e. dissimilarity)
    per ITK/ANTs convention. E.g. using Correlation metric, the similarity
    of an image with itself returns -1.

    ANTsR function: `imageSimilarity`

    Arguments
    ---------
    fixed : ANTsImage
        the fixed image

    moving : ANTsImage
        the moving image

    metric_type : string
        image metric to calculate
            MeanSquares
            Correlation
            ANTSNeighborhoodCorrelation
            MattesMutualInformation
            JointHistogramMutualInformation
            Demons

    fixed_mask : ANTsImage (optional)
        mask for the fixed image

    moving_mask : ANTsImage (optional)
        mask for the moving image

    sampling_strategy : string (optional)
        sampling strategy, default is full sampling
            None (Full sampling)
            random
            regular

    sampling_percentage : scalar
        percentage of data to sample when calculating metric
        Must be between 0 and 1

    Returns
    -------
    scalar

    Example
    -------
    >>> import ants
    >>> x = ants.image_read(ants.get_ants_data('r16'))
    >>> y = ants.image_read(ants.get_ants_data('r30'))
    >>> metric = ants.image_similarity(x,y,metric_type='MeanSquares')
    """
    metric = ants.create_ants_metric(fixed_image, moving_image, metric_type, fixed_mask,
                        moving_mask, sampling_strategy, sampling_percentage)
    return metric.get_value()




__all__ = ["label_stats"]

import pandas as pd

from ants.internal import get_lib_fn
from ants.decorators import image_method

@image_method
def label_stats(image, label_image):
    """
    Get label statistics from image

    ANTsR function: `labelStats`

    Arguments
    ---------
    image : ANTsImage
        Image from which statistics will be calculated

    label_image : ANTsImage
        Label image

    Returns
    -------
    ndarray ?

    Example
    -------
    >>> import ants
    >>> image = ants.image_read( ants.get_ants_data('r16') , 2 )
    >>> image = ants.resample_image( image, (64,64), 1, 0 )
    >>> mask = ants.get_mask(image)
    >>> segs1 = ants.kmeans_segmentation( image, 3 )
    >>> stats = ants.label_stats(image, segs1['segmentation'])
    """
    image_float = image.clone("float")
    label_image_int = label_image.clone("unsigned int")

    libfn = get_lib_fn("labelStats%iD" % image.dimension)
    df = libfn(image_float.pointer, label_image_int.pointer)
    df = pd.DataFrame(df)
    df.sort_values(by=["LabelValue"], inplace=True)
    return df




__all__ = ['integrate_velocity_field']

import ants
from ants.internal import get_lib_fn


def integrate_velocity_field(velocity_field,
                             lower_integration_bound=0.0,
                             upper_integration_bound=1.0,
                             number_of_integration_steps=10):
    """
    Integrate velocity field.

    Arguments
    ---------
    velocity_field : ANTsImage velocity field
        time-varying displacement field

    lower_integration_bound: float
        Lower time bound for integration in [0, 1]

    upper_integration_bound: float
        Upper time bound for integration in [0, 1]

    number_of_integation_steps: integer
        Number of integration steps used in the Runge-Kutta solution

    Example
    -------
    >>> import ants
    >>> fi = ants.image_read( ants.get_data( "r16" ) )
    >>> mi = ants.image_read( ants.get_data( "r27" ) )
    >>> reg = ants.registration(fi, mi, "TV[2]")
    >>> velocity_field = ants.image_read(reg['velocityfield'][0])
    >>> field = ants.integrate_velocity_field(velocity_field, 0.0, 1.0, 10) 
    >>> temp=ants.apply_ants_transform_to_image( 
	    ants.transform_from_displacement_field( field ), mi, fi )
    """

    libfn = get_lib_fn('integrateVelocityFieldD%i' % (velocity_field.dimension-1))
    integrated_field = libfn(velocity_field.pointer, lower_integration_bound,
        upper_integration_bound, number_of_integration_steps)

    new_image = ants.from_pointer(integrated_field).clone('float')
    return new_image






 

__all__ = ['crop_image', 
           'crop_indices',
           'decrop_image']


import ants
from ants.decorators import image_method
from ants.internal import get_lib_fn

@image_method
def crop_image(image, label_image=None, label=1):
    """
    Use a label image to crop a smaller ANTsImage from within a larger ANTsImage

    ANTsR function: `cropImage`
    
    Arguments
    ---------
    image : ANTsImage  
        image to crop
    
    label_image : ANTsImage
        image with label values. If not supplied, estimated from data.
    
    label : integer   
        the label value to use

    Returns
    -------
    ANTsImage

    Example
    -------
    >>> import ants
    >>> fi = ants.image_read( ants.get_ants_data('r16') )
    >>> cropped = ants.crop_image(fi)
    >>> fi2 = ants.merge_channels([fi,fi])
    >>> cropped2 = ants.crop_image(fi2)
    >>> cropped = ants.crop_image(fi, fi, 100 )
    """
    if image.has_components:
        return ants.merge_channels([crop_image(img, label_image, label) for img in ants.split_channels(image)],
                                   channels_first=image.channels_first)
        
    inpixeltype = image.pixeltype
    ndim = image.dimension
    if image.pixeltype != 'float':
        image = image.clone('float')

    if label_image is None:
        label_image = ants.get_mask(image)

    if label_image.pixeltype != 'float':
        label_image = label_image.clone('float')

    libfn = get_lib_fn('cropImage')
    itkimage = libfn(image.pointer, label_image.pointer, label, 0, [], [])
    return ants.from_pointer(itkimage).clone(inpixeltype)


@image_method
def crop_indices(image, lowerind, upperind):
    """
    Create a proper ANTsImage sub-image by indexing the image with indices. 
    This is similar to but different from array sub-setting in that 
    the resulting sub-image can be decropped back into its place without 
    having to store its original index locations explicitly.
    
    ANTsR function: `cropIndices`

    Arguments
    ---------
    image : ANTsImage  
        image to crop
    
    lowerind : list/tuple of integers  
        vector of lower index, should be length image dimensionality
    
    upperind : list/tuple of integers
        vector of upper index, should be length image dimensionality
    
    Returns
    -------
    ANTsImage

    Example
    -------
    >>> import ants
    >>> fi = ants.image_read( ants.get_ants_data("r16"))
    >>> cropped = ants.crop_indices( fi, (10,10), (100,100) )
    >>> cropped = ants.smooth_image( cropped, 5 )
    >>> decropped = ants.decrop_image( cropped, fi )
    """
    if image.has_components:
        return ants.merge_channels([crop_indices(img, lowerind, upperind) for img in ants.split_channels(image)],
                                   channels_first=image.channels_first)
        
    inpixeltype = 'float'
    if image.pixeltype != 'float':
        inpixeltype = image.pixeltype
        image = image.clone('float')

    if (image.dimension != len(lowerind)) or (image.dimension != len(upperind)):
        raise ValueError('image dimensionality and index length must match')

    libfn = get_lib_fn('cropImage')
    itkimage = libfn(image.pointer, image.pointer, 1, 2, lowerind, upperind)
    ants_image = ants.from_pointer(itkimage)
    if inpixeltype != 'float':
        ants_image = ants_image.clone(inpixeltype)
    return ants_image

@image_method
def decrop_image(cropped_image, full_image):
    """
    The inverse function for `ants.crop_image`

    ANTsR function: `decropImage`
    
    Arguments
    ---------
    cropped_image : ANTsImage
        cropped image

    full_image : ANTsImage
        image in which the cropped image will be put back

    Returns
    -------
    ANTsImage

    Example
    -------
    >>> import ants
    >>> fi = ants.image_read(ants.get_ants_data('r16'))
    >>> mask = ants.get_mask(fi)
    >>> cropped = ants.crop_image(fi, mask, 1)
    >>> cropped = ants.smooth_image(cropped, 1)
    >>> decropped = ants.decrop_image(cropped, fi)
    """
    inpixeltype = 'float'
    if cropped_image.pixeltype != 'float':
        inpixeltype= cropped_image.pixeltype
        cropped_image = cropped_image.clone('float')
    if full_image.pixeltype != 'float':
        full_image = full_image.clone('float')
    
    libfn = get_lib_fn('cropImage')
    itkimage = libfn(cropped_image.pointer, full_image.pointer, 1, 1, [], [])
    ants_image = ants.from_pointer(itkimage)
    if inpixeltype != 'float':
        ants_image = ants_image.clone(inpixeltype)

    return ants_image




from .image_to_cluster_images import image_to_cluster_images
from .label_clusters import label_clusters
from .label_geometry_measures import label_geometry_measures
from .label_image_centroids import label_image_centroids
from .label_overlap_measures import label_overlap_measures
from .label_stats import label_stats
from .labels_to_matrix import labels_to_matrix
from .make_points_image import make_points_image
from .multi_label_morphology import multi_label_morphology





__all__ = ['rgb_to_vector', 
           'vector_to_rgb', 
           'scalar_to_rgb']

import os
from tempfile import mktemp

import numpy as np

import ants
from ants.internal import get_lib_fn, process_arguments
from ants.decorators import image_method

def scalar_to_rgb(image, mask=None, filename=None, cmap='red', custom_colormap_file=None, 
                  min_input=None, max_input=None, min_rgb_output=None, max_rgb_output=None,
                  vtk_lookup_table=None):
    """
    Usage: ConvertScalarImageToRGB imageDimension inputImage outputImage mask colormap 
    [customColormapFile] [minimumInput] [maximumInput] [minimumRGBOutput=0] 
    [maximumRGBOutput=255] <vtkLookupTable>
    Possible colormaps: grey, red, green, blue, copper, jet, hsv, spring, summer, autumn, winter, hot, cool, overunder, custom
    
    Example
    -------
    >>> import ants
    >>> img = ants.image_read(ants.get_data('r16'))
    >>> img_color = ants.scalar_to_rgb(img, cmap='jet')
    """
    raise Exception('This function is currently not supported.')

@image_method
def rgb_to_vector(image):
    """
    Convert an RGB ANTsImage to a Vector ANTsImage

    Arguments
    ---------
    image : ANTsImage
        RGB image to be converted

    Returns
    -------
    ANTsImage

    Example
    -------
    >>> import ants
    >>> mni = ants.image_read(ants.get_data('mni'))
    >>> mni_rgb = ants.scalar_to_rgb(mni)
    >>> mni_vector = mni.rgb_to_vector()
    >>> mni_rgb2 = mni.vector_to_rgb()
    """
    if image.pixeltype != 'unsigned char':
        image = image.clone('unsigned char')
    idim = image.dimension
    libfn = get_lib_fn('RgbToVector%i' % idim)
    new_ptr = libfn(image.pointer)
    new_img = ants.from_pointer(new_ptr)
    return new_img

@image_method
def vector_to_rgb(image):
    """
    Convert an Vector ANTsImage to a RGB ANTsImage

    Arguments
    ---------
    image : ANTsImage
        RGB image to be converted

    Returns
    -------
    ANTsImage

    Example
    -------
    >>> import ants
    >>> img = ants.image_read(ants.get_data('r16'), pixeltype='unsigned char')
    >>> img_rgb = ants.scalar_to_rgb(img.clone())
    >>> img_vec = img_rgb.rgb_to_vector()
    >>> img_rgb2 = img_vec.vector_to_rgb()
    """
    if image.pixeltype != 'unsigned char':
        image = image.clone('unsigned char')
    idim = image.dimension
    libfn = get_lib_fn('VectorToRgb%i' % idim)
    new_ptr = libfn(image.pointer)
    new_img = ants.from_pointer(new_ptr)
    return new_img




__all__ = ["fit_thin_plate_spline_displacement_field"]

import numpy as np

import ants
from ants.internal import get_lib_fn


def fit_thin_plate_spline_displacement_field(displacement_origins=None,
                                             displacements=None,
                                             origin=None,
                                             spacing=None,
                                             size=None,
                                             direction=None):

    """
    Fit a thin-plate spline object to a a set of points with associated displacements.  
    This is basically a wrapper for the ITK filter

    https://itk.org/Doxygen/html/itkThinPlateSplineKernelTransform_8h.html

    ANTsR function: `fitThinPlateSplineToDisplacementField`

    Arguments
    ---------

    displacement_origins : 2-D numpy array
        Matrix (number_of_points x dimension) defining the origins of the input
        displacement points.  Default = None.

    displacements : 2-D numpy array
        Matrix (number_of_points x dimension) defining the displacements of the input
        displacement points.  Default = None.

    origin : n-D tuple
        Defines the physical origin of the B-spline object.

    spacing : n-D tuple
        Defines the physical spacing of the B-spline object.

    size : n-D tuple
       Defines the size (length) of the spline object.  Note that the length of the
       spline object in dimension d is defined as spacing[d] * size[d]-1.

    direction : 2-D numpy array
       Booleans defining whether or not the corresponding parametric dimension is
       closed (e.g., closed loop).  Default = None.

    Returns
    -------
    Returns an ANTsImage.

    Example
    -------
    >>> import ants
    >>> import numpy as np
    >>> points = np.array([[-50, -50]])
    >>> deltas = np.array([[10, 10]])
    >>> tps_field = ants.fit_thin_plate_spline_displacement_field(
    >>>   displacement_origins=points, displacements=deltas,
    >>>   origin=[0.0, 0.0], spacing=[1.0, 1.0], size=[100, 100],
    >>>   direction=np.array([[-1, 0], [0, -1]]))
    """

    dimensionality = displacement_origins.shape[1]
    if displacements.shape[1] != dimensionality:
        raise ValueError("Dimensionality between origins and displacements does not match.")

    if displacement_origins is None or displacement_origins is None:
        raise ValueError("Missing input.  Input point set (origins + displacements) needs to be specified." )

    if origin is not None and len(origin) != dimensionality:
        raise ValueError("Origin is not of length dimensionality.")

    if spacing is not None and len(spacing) != dimensionality:
        raise ValueError("Spacing is not of length dimensionality.")

    if size is not None and len(size) != dimensionality:
        raise ValueError("Size is not of length dimensionality.")

    if direction is not None and (direction.shape[0] != dimensionality and direction.shape[1] != dimensionality):
        raise ValueError("Direction is not of shape dimensionality x dimensionality.")

    # It would seem that pybind11 doesn't really play nicely when the
    # arguments are 'None'

    if origin is None:
        origin = np.empty(0)

    if spacing is None:
        spacing = np.empty(0)

    if size is None:
        size = np.empty(0)

    if direction is None:
        direction = np.empty((0, 0))

    tps_field = None
    libfn = get_lib_fn("fitThinPlateSplineDisplacementFieldToScatteredDataD%i" % (dimensionality))
    tps_field = libfn(displacement_origins, displacements, origin, spacing, size, direction)

    tps_displacement_field = ants.from_pointer(tps_field).clone('float')
    return tps_displacement_field




__all__ = ["hausdorff_distance"]

from ants.decorators import image_method
from ants.internal import get_lib_fn

@image_method
def hausdorff_distance(image1, image2):
    """
    Get Hausdorff distance between non-zero pixels in two images

    ANTsR function: `hausdorffDistance`

    Arguments
    ---------
    source image : ANTsImage
        Source image

    target_image : ANTsImage
        Target image

    Returns
    -------
    data frame with "Distance" and "AverageDistance"

    Example
    -------
    >>> import ants
    >>> r16 = ants.image_read( ants.get_ants_data('r16') )
    >>> r64 = ants.image_read( ants.get_ants_data('r64') )
    >>> s16 = ants.kmeans_segmentation( r16, 3 )['segmentation']
    >>> s64 = ants.kmeans_segmentation( r64, 3 )['segmentation']
    >>> stats = ants.hausdorff_distance(s16, s64)
    """
    image1_int = image1.clone("unsigned int")
    image2_int = image2.clone("unsigned int")

    libfn = get_lib_fn("hausdorffDistance%iD" % image1_int.dimension)
    d = libfn(image1_int.pointer, image2_int.pointer)

    return d



__all__ = ["n3_bias_field_correction",
           "n3_bias_field_correction2", 
           "n4_bias_field_correction",
           "abp_n4"]

import ants
from ants.decorators import image_method
from ants.internal import get_lib_fn, get_pointer_string, process_arguments

@image_method
def n3_bias_field_correction(image, downsample_factor=3):
    """
    N3 Bias Field Correction

    ANTsR function: `n3BiasFieldCorrection`

    Arguments
    ---------
    image : ANTsImage
        image to be bias corrected

    downsample_factor : scalar
        how much to downsample image before performing bias correction

    Returns
    -------
    ANTsImage

    Example
    -------
    >>> import ants
    >>> image = ants.image_read( ants.get_ants_data('r16') )
    >>> image_n3 = ants.n3_bias_field_correction(image)
    """
    outimage = image.clone()
    args = [image.dimension, image, outimage, downsample_factor]
    processed_args = process_arguments(args)
    libfn = get_lib_fn("N3BiasFieldCorrection")
    libfn(processed_args)
    return outimage

@image_method
def n3_bias_field_correction2(
    image,
    mask=None,
    rescale_intensities=False,
    shrink_factor=4,
    convergence={"iters": 50, "tol": 1e-7},
    spline_param=None,
    number_of_fitting_levels=4,
    return_bias_field=False,
    verbose=False,
    weight_mask=None,
):
    """
    N3 Bias Field Correction

    ANTsR function: `n3BiasFieldCorrection2`

    Arguments
    ---------
    image : ANTsImage
        image to bias correct

    mask : ANTsImage
        Input mask.  If not specified, the entire image is used.

    rescale_intensities : boolean
        At each iteration, a new intensity mapping is
        calculated and applied but there is nothing which constrains the new
        intensity range to be within certain values. The result is that the
        range can "drift" from the original at each iteration. This option
        rescales to the [min,max] range of the original image intensities within
        the user-specified mask. A mask is required to perform rescaling.  Default
        is False in ANTsR/ANTsPy but True in ANTs.

    shrink_factor : scalar
        Shrink factor for multi-resolution correction, typically integer less than 4

    convergence : dict w/ keys `iters` and `tol`
        iters : maximum number of iterations
        tol : the convergence tolerance.  Default tolerance is 1e-7 in ANTsR/ANTsPy but 0.0 in ANTs.

    spline_param : float or vector Parameter controlling number of control
        points in spline. Either single value, indicating the spacing in each
        direction, or vector with one entry per dimension of image, indicating
        the mesh size. If None, defaults to mesh size of 1 in all dimensions.

    number_of_fitting_levels : integer
        Number of fitting levels per iteration.

    return_bias_field : boolean
        Return bias field instead of bias corrected image.

    verbose : boolean
        enables verbose output.

    weight_mask : ANTsImage (optional)
        antsImage of weight mask

    Returns
    -------
    ANTsImage

    Example
    -------
    >>> image = ants.image_read( ants.get_ants_data('r16') )
    >>> image_n3 = ants.n3_bias_field_correction2(image)
    """
    if image.pixeltype != "float":
        image = image.clone("float")
    iters = convergence["iters"]
    tol = convergence["tol"]
    if mask is None:
        mask = image * 0 + 1
    if spline_param is None:
        spline_param = [1] * image.dimension

    N3_CONVERGENCE_1 = "[%i,%.10f]" % (iters, tol)
    N3_SHRINK_FACTOR_1 = str(shrink_factor)
    if (not isinstance(spline_param, (list, tuple))) or (len(spline_param) == 1):
        N3_BSPLINE_PARAMS = "[%i,%i]" % (spline_param, number_of_fitting_levels)
    elif (isinstance(spline_param, (list, tuple))) and (
        len(spline_param) == image.dimension
    ):
        N3_BSPLINE_PARAMS = "[%s,%i]" % (("x".join([str(sp) for sp in spline_param])), number_of_fitting_levels)
    else:
        raise ValueError(
            "Length of splineParam must either be 1 or dimensionality of image"
        )

    if weight_mask is not None:
        if not ants.is_image(weight_mask):
            raise ValueError("Weight Image must be an antsImage")

    outimage = image.clone("float")
    outbiasfield = image.clone("float")
    i = get_pointer_string(outimage)
    b = get_pointer_string(outbiasfield)
    output = "[%s,%s]" % (i, b)

    kwargs = {
        "d": outimage.dimension,
        "i": image,
        "w": weight_mask,
        "s": N3_SHRINK_FACTOR_1,
        "c": N3_CONVERGENCE_1,
        "b": N3_BSPLINE_PARAMS,
        "x": mask,
        "r": int(rescale_intensities),
        "o": output,
        "v": int(verbose),
    }

    processed_args = process_arguments(kwargs)
    libfn = get_lib_fn("N3BiasFieldCorrection")
    libfn(processed_args)
    if return_bias_field == True:
        return outbiasfield
    else:
        return outimage

@image_method
def n4_bias_field_correction(
    image,
    mask=None,
    rescale_intensities=False,
    shrink_factor=4,
    convergence={"iters": [50, 50, 50, 50], "tol": 1e-7},
    spline_param=None,
    return_bias_field=False,
    verbose=False,
    weight_mask=None,
):
    """
    N4 Bias Field Correction

    ANTsR function: `n4BiasFieldCorrection`

    Arguments
    ---------
    image : ANTsImage
        image to bias correct

    mask : ANTsImage
        Input mask.  If not specified, the entire image is used.

    rescale_intensities : boolean
        At each iteration, a new intensity mapping is
        calculated and applied but there is nothing which constrains the new
        intensity range to be within certain values. The result is that the
        range can "drift" from the original at each iteration. This option
        rescales to the [min,max] range of the original image intensities within
        the user-specified mask. A mask is required to perform rescaling.  Default
        is False in ANTsR/ANTsPy but True in ANTs.

    shrink_factor : scalar
        Shrink factor for multi-resolution correction, typically integer less than 4

    convergence : dict w/ keys `iters` and `tol`
        iters : vector of maximum number of iterations for each level
        tol : the convergence tolerance.  Default tolerance is 1e-7 in ANTsR/ANTsPy but 0.0 in ANTs.

    spline_param : float or vector
        Parameter controlling number of control points in spline. Either single value,
        indicating the spacing in each direction, or vector with one entry per
        dimension of image, indicating the mesh size. If None, defaults to mesh size of 1 in all
        dimensions.

    return_bias_field : boolean
        Return bias field instead of bias corrected image.

    verbose : boolean
        enables verbose output.

    weight_mask : ANTsImage (optional)
        antsImage of weight mask

    Returns
    -------
    ANTsImage

    Example
    -------
    >>> image = ants.image_read( ants.get_ants_data('r16') )
    >>> image_n4 = ants.n4_bias_field_correction(image)
    """
    if image.pixeltype != "float":
        image = image.clone("float")
    iters = convergence["iters"]
    tol = convergence["tol"]

    if mask is None:
        mask = image * 0 + 1
    if spline_param is None:
        spline_param = [1] * image.dimension

    N4_CONVERGENCE_1 = "[%s, %.10f]" % ("x".join([str(it) for it in iters]), tol)
    N4_SHRINK_FACTOR_1 = str(shrink_factor)
    if (not isinstance(spline_param, (list, tuple))) or (len(spline_param) == 1):
        N4_BSPLINE_PARAMS = "[%i]" % spline_param
    elif (isinstance(spline_param, (list, tuple))) and (
        len(spline_param) == image.dimension
    ):
        N4_BSPLINE_PARAMS = "[%s]" % ("x".join([str(sp) for sp in spline_param]))
    else:
        raise ValueError(
            "Length of splineParam must either be 1 or dimensionality of image"
        )

    if weight_mask is not None:
        if not ants.is_image(weight_mask):
            raise ValueError("Weight Image must be an antsImage")

    outimage = image.clone("float")
    outbiasfield = image.clone("float")
    i = get_pointer_string(outimage)
    b = get_pointer_string(outbiasfield)
    output = "[%s,%s]" % (i, b)

    kwargs = {
        "d": outimage.dimension,
        "i": image,
        "w": weight_mask,
        "s": N4_SHRINK_FACTOR_1,
        "c": N4_CONVERGENCE_1,
        "b": N4_BSPLINE_PARAMS,
        "x": mask,
        "r": int(rescale_intensities),
        "o": output,
        "v": int(verbose),
    }

    processed_args = process_arguments(kwargs)
    libfn = get_lib_fn("N4BiasFieldCorrection")
    libfn(processed_args)
    if return_bias_field == True:
        return outbiasfield
    else:
        return outimage

@image_method
def abp_n4(image, intensity_truncation=(0.025, 0.975, 256), mask=None, usen3=False):
    """
    Truncate outlier intensities and bias correct with the N4 algorithm.

    ANTsR function: `abpN4`

    Arguments
    ---------
    image : ANTsImage
        image to correct and truncate

    intensity_truncation : 3-tuple
        quantiles for intensity truncation

    mask : ANTsImage (optional)
        mask for bias correction

    usen3 : boolean
        if True, use N3 bias correction instead of N4

    Returns
    -------
    ANTsImage

    Example
    -------
    >>> import ants
    >>> image = ants.image_read(ants.get_ants_data('r16'))
    >>> image2 = ants.abp_n4(image)
    """
    if (not isinstance(intensity_truncation, (list, tuple))) or (
        len(intensity_truncation) != 3
    ):
        raise ValueError("intensity_truncation must be list/tuple with 3 values")
    outimage = ants.iMath(
        image,
        "TruncateIntensity",
        intensity_truncation[0],
        intensity_truncation[1],
        intensity_truncation[2],
    )
    if usen3 == True:
        outimage = n3_bias_field_correction(outimage, 4)
        outimage = n3_bias_field_correction(outimage, 2)
        return outimage
    else:
        outimage = n4_bias_field_correction(outimage, mask)
        return outimage




__all__ = ['fuzzy_spatial_cmeans_segmentation']

import numpy as np

import ants

def fuzzy_spatial_cmeans_segmentation(image,
                                      mask=None,
                                      number_of_clusters=4,
                                      m=2,
                                      p=1,
                                      q=1,
                                      radius=2,
                                      max_number_of_iterations=20,
                                      convergence_threshold=0.02,
                                      verbose=False):
    """
    Fuzzy spatial c-means for image segmentation.

    Image segmentation using fuzzy spatial c-means as described in

    Chuang et al., Fuzzy c-means clustering with spatial information for image
    segmentation.  CMIG: 30:9-15, 2006.

    Arguments
    ---------
    image : ANTsImage
        Input image.

    mask : ANTsImage
        Optional mask image.

    number_of_clusters : integer
        Number of segmentation clusters.

    m : float
        Fuzziness parameter (default=2).

    p : float
        Membership importance parameter (default=1).

    q : float
        Spatial constraint importance parameter (default=1).
        q = 0 is equivalent to conventional fuzzy c-means.

    radius : integer or tuple
        Neighborhood radius (scalar or array) for spatial constraint.

    max_number_of_iterations : integer
        Iteration limit (default=20).

    convergence_threshold : float
        Convergence between iterations is measured using the Dice coefficient
        (default=0.02).

    varbose : boolean
        Print progress.

    Returns
    -------
    dictionary containing ANTsImage and probability images

    Example
    -------
    >>> import ants
    >>> image = ants.image_read(ants.get_ants_data('r16'))
    >>> mask = ants.get_mask(image)
    >>> fuzzy = ants.fuzzy_spatial_cmeans_segmentation(image, mask, number_of_clusters=3)
    """

    if mask is None:
        mask = ants.image_clone(image) * 0 + 1

    x = image[mask != 0]

    # This is a hack because the order of the voxels is not the same for the following
    # two operations
    #        x = image[mask != 0]
    #        x = ants.get_neighborhood_in_mask(image, mask)

    mask_perm = ants.from_numpy(np.transpose(mask.numpy()))

    v = np.linspace(0, 1, num=(number_of_clusters + 2))[1:(number_of_clusters+1)]
    v = v * (x.max() - x.min()) + x.min()
    cc = len(v)

    if verbose == True:
        print("Initial cluster centers: ", v)

    xx = np.zeros((cc, len(x)))
    for i in range(cc):
        xx[i,:] = x

    if isinstance(radius, int):
        radius = tuple(np.zeros((image.dimension,), dtype=int) + radius)

    segmentation = ants.image_clone(image) * 0
    probability_images = None

    np.seterr(divide='ignore', invalid='ignore')

    iter = 0
    dice_value = 0
    while iter < max_number_of_iterations and dice_value < 1.0 - convergence_threshold:

        # Update membership values

        xv = np.zeros((cc, len(x)))
        for k in range(cc):
            xv[k,:] = abs(x - v[k])

        u = np.zeros((xv.shape[0], xv.shape[1]))
        for i in range(cc):
            n = xv[i,:]

            d = n * 0
            for k in range(cc):
                d += (n / xv[k,:]) ** (2 / (m - 1))

            u[i,:] = 1 / d
        u = np.nan_to_num(u, nan=1.0)

        # Update cluster centers

        v = np.nansum((u ** m) * xx, axis=1) / np.nansum((u ** m), axis=1)

        if verbose == True:
            print("Updated cluster centers: ", v)

        # Spatial function

        h = np.zeros((u.shape[0], u.shape[1]))
        for i in range(cc):
            u_image = ants.image_clone(image) * 0
            u_image[mask != 0] = u[i,:]
            u_image_perm = ants.from_numpy(np.transpose(u_image.numpy()))
            u_neighborhoods = ants.get_neighborhood_in_mask(u_image_perm, mask_perm, radius)
            h[i,:] = np.nansum(u_neighborhoods, axis=0)

        # u prime

        d = np.zeros((u.shape[1],))
        for k in range(cc):
            d += (u[k,:] ** p) * (h[k,:] ** q)

        probability_images = []
        uprime = np.zeros((u.shape[0], u.shape[1]))
        for i in range(cc):
            uprime[i,:] = ((u[i,:] ** p) * (h[i,:] ** q)) / d
            uprime_image = ants.image_clone(image) * 0
            uprime_image[mask != 0] = uprime[i,:]
            probability_images.append(uprime_image)

        tmp_segmentation = ants.image_clone(image) * 0
        tmp_segmentation[mask != 0] = np.argmax(uprime, axis=0) + 1

        dice_value = ants.label_overlap_measures(segmentation, tmp_segmentation)['MeanOverlap'][0]
        iter = iter + 1

        if verbose == True:
            print("Iteration ", iter, " (out of ", max_number_of_iterations, "):  ",
                "Dice overlap = ", dice_value, sep = "")

        segmentation = tmp_segmentation

    return_dict = {'segmentation_image' : segmentation,
                   'probability_images' : probability_images}
    return(return_dict)


__all__ = ["denoise_image"]


from ants.decorators import image_method
from ants.internal import get_lib_fn, process_arguments

@image_method
def denoise_image(image, mask=None, shrink_factor=1, p=1, r=2, noise_model="Rician", v=0):
    """
    Denoise an image using a spatially adaptive filter originally described in
    J. V. Manjon, P. Coupe, Luis Marti-Bonmati, D. L. Collins, and M. Robles.
    Adaptive Non-Local Means Denoising of MR Images With Spatially Varying
    Noise Levels, Journal of Magnetic Resonance Imaging, 31:192-203, June 2010.

    ANTsR function: `denoiseImage`

    Arguments
    ---------
    image : ANTsImage
        scalar image to denoise.

    mask : ANTsImage
        to limit the denoise region.

    shrink_factor : scalar
        downsampling level performed within the algorithm.

    p : integer or character of format '2x2' where the x separates vector entries
        patch radius for local sample.

    r : integer or character of format '2x2' where the x separates vector entries
        search radius from which to choose extra local samples.

    noise_model : string
        'Rician' or 'Gaussian'

    Returns
    -------
    ANTsImage

    Example
    -------
    >>> import ants
    >>> import numpy as np
    >>> image = ants.image_read(ants.get_ants_data('r16'))
    >>> # add fairly large salt and pepper noise
    >>> imagenoise = image + np.random.randn(*image.shape).astype('float32')*5
    >>> imagedenoise = ants.denoise_image(imagenoise, ants.get_mask(image))
    """
    inpixeltype = image.pixeltype
    outimage = image.clone("float")

    mydim = image.dimension

    if mask is None:
        myargs = {
            "d": mydim,
            "i": image,
            "n": noise_model,
            "s": int(shrink_factor),
            "p": p,
            "r": r,
            "o": outimage,
            "v": v,
        }
    else:
        myargs = {
            "d": mydim,
            "i": image,
            "n": noise_model,
            "x": mask.clone("unsigned char"),
            "s": int(shrink_factor),
            "p": p,
            "r": r,
            "o": outimage,
            "v": v,
        }

    processed_args = process_arguments(myargs)
    libfn = get_lib_fn("DenoiseImage")
    libfn(processed_args)
    return outimage.clone(inpixeltype)




__all__ = ['image_to_cluster_images']

import numpy as np

import ants
from ants.decorators import image_method

@image_method
def image_to_cluster_images(image, min_cluster_size=50, min_thresh=1e-06, max_thresh=1):
    """
    Converts an image to several independent images.

    Produces a unique image for each connected 
    component 1 through N of size > min_cluster_size

    ANTsR function: `image2ClusterImages`

    Arguments
    ---------
    image : ANTsImage
        input image
    min_cluster_size : integer
        throw away clusters smaller than this value
    min_thresh : scalar
        threshold to a statistical map
    max_thresh : scalar
        threshold to a statistical map

    Returns
    -------
    list of ANTsImage types

    Example
    -------
    >>> import ants
    >>> image = ants.image_read(ants.get_ants_data('r16'))
    >>> image = ants.threshold_image(image, 1, 1e15)
    >>> image_cluster_list = ants.image_to_cluster_images(image)
    """
    if not ants.is_image(image):
        raise ValueError('image must be ANTsImage type')

    clust = ants.label_clusters(image, min_cluster_size, min_thresh, max_thresh)
    labs = np.unique(clust[clust > 0])

    clustlist = []
    for i in range(len(labs)):
        labimage = image.clone()
        labimage[clust != labs[i]] = 0
        clustlist.append(labimage)
    return clustlist




"""
Functions for plotting ants images
"""


__all__ = [
    "plot_directory"
]

import fnmatch
import math
import os
import warnings

from matplotlib import gridspec
import matplotlib.pyplot as plt
import matplotlib.patheffects as path_effects
import matplotlib.lines as mlines
import matplotlib.patches as patches
import matplotlib.mlab as mlab
import matplotlib.animation as animation
from mpl_toolkits.axes_grid1.inset_locator import inset_axes


import numpy as np

import ants


def plot_directory(
    directory,
    recursive=False,
    regex="*",
    save_prefix="",
    save_suffix="",
    axis=None,
    **kwargs
):
    """
    Create and save an ANTsPy plot for every image matching a given regular
    expression in a directory, optionally recursively. This is a good function
    for quick visualize exploration of all of images in a directory

    ANTsR function: N/A

    Arguments
    ---------
    directory : string
        directory in which to search for images and plot them

    recursive : boolean
        If true, this function will search through all directories under
        the given directory recursively to make plots.
        If false, this function will only create plots for images in the
        given directory

    regex : string
        regular expression used to filter out certain filenames or suffixes

    save_prefix : string
        sub-string that will be appended to the beginning of all saved plot filenames.
        Default is to add nothing.

    save_suffix : string
        sub-string that will be appended to the end of all saved plot filenames.
        Default is add nothing.

    kwargs : keyword arguments
        any additional arguments to pass onto the `ants.plot` function.
        e.g. overlay, alpha, cmap, etc. See `ants.plot` for more options.

    Example
    -------
    >>> import ants
    >>> ants.plot_directory(directory='~/desktop/testdir',
                            recursive=False, regex='*')
    """

    def has_acceptable_suffix(fname):
        suffixes = {".nii.gz"}
        return sum([fname.endswith(sx) for sx in suffixes]) > 0

    if directory.startswith("~"):
        directory = os.path.expanduser(directory)

    if not os.path.isdir(directory):
        raise ValueError("directory %s does not exist!" % directory)

    for root, dirnames, fnames in os.walk(directory):
        for fname in fnames:
            if fnmatch.fnmatch(fname, regex) and has_acceptable_suffix(fname):
                load_fname = os.path.join(root, fname)
                fname = fname.replace(".".join(fname.split(".")[1:]), "png")
                fname = fname.replace(".png", "%s.png" % save_suffix)
                fname = "%s%s" % (save_prefix, fname)
                save_fname = os.path.join(root, fname)
                img = ants.image_read(load_fname)

                if axis is None:
                    axis_range = [i for i in range(img.dimension)]
                else:
                    axis_range = axis if isinstance(axis, (list, tuple)) else [axis]

                if img.dimension > 2:
                    for axis_idx in axis_range:
                        filename = save_fname.replace(".png", "_axis%i.png" % axis_idx)
                        ncol = int(math.sqrt(img.shape[axis_idx]))
                        ants.plot(
                            img,
                            axis=axis_idx,
                            nslices=img.shape[axis_idx],
                            ncol=ncol,
                            filename=filename,
                            **kwargs
                        )
                else:
                    filename = save_fname
                    ants.plot(img, filename=filename, **kwargs)





__all__ = ['symmetrize_image']

from tempfile import mktemp

import ants
from ants.decorators import image_method

@image_method
def symmetrize_image(image):
    """
    Use registration and reflection to make an image symmetric

    ANTsR function: N/A

    Arguments
    ---------
    image : ANTsImage
        image to make symmetric

    Returns
    -------
    ANTsImage

    Example
    -------
    >>> import ants
    >>> image = ants.image_read( ants.get_ants_data('r16') , 'float')
    >>> simage = ants.symimage(image)
    """
    imager = ants.reflect_image(image, axis=0)
    imageavg = imager * 0.5 + image
    for i in range(5):
        w1 = ants.registration(imageavg, image, type_of_transform='SyN')
        w2 = ants.registration(imageavg, imager, type_of_transform='SyN')
        
        xavg = w1['warpedmovout']*0.5 + w2['warpedmovout']*0.5
        nada1 = ants.apply_transforms(image, image, w1['fwdtransforms'], compose=w1['fwdtransforms'][0])
        nada2 = ants.apply_transforms(image, image, w2['fwdtransforms'], compose=w2['fwdtransforms'][0])

        wavg = (ants.image_read(nada1) + ants.image_read(nada2)) * (-0.5)
        wavgfn = mktemp(suffix='.nii.gz')
        ants.image_write(wavg, wavgfn)
        xavg = ants.apply_transforms(image, imageavg, wavgfn)

    return xavg





__all__ = ['ndimage_to_list',
           'list_to_ndimage']


import numpy as np

import ants
from ants.decorators import image_method

@image_method
def list_to_ndimage( image, image_list ):
    """
    Merge list of multiple scalar ANTsImage types of dimension into one
    ANTsImage of dimension plus one

    ANTsR function: `mergeListToNDImage`

    Arguments
    ---------
    image : target image space
    image_list : list/tuple of ANTsImage types
        scalar images to merge into target image space

    Returns
    -------
    ANTsImage

    Example
    -------
    >>> import ants
    >>> image = ants.image_read(ants.get_ants_data('r16'))
    >>> image2 = ants.image_read(ants.get_ants_data('r16'))
    >>> imageTar = ants.make_image( ( *image2.shape, 2 ) )
    >>> image3 = ants.list_to_ndimage( imageTar, [image,image2])
    >>> image3.dimension == 3
    """
    inpixeltype = image_list[0].pixeltype
    dimension = image_list[0].dimension
    components = len(image_list)

    for imageL in image_list:
        if not ants.is_image(imageL):
            raise ValueError('list may only contain ANTsImage objects')
        if image.pixeltype != inpixeltype:
            raise ValueError('all images must have the same pixeltype')

    dimensionout = ( *image_list[0].shape, len( image_list )  )
    newImage = ants.make_image(
      dimensionout,
      spacing = ants.get_spacing( image ),
      origin = ants.get_origin( image ),
      direction = ants.get_direction( image ),
      pixeltype = inpixeltype
      )
    # FIXME - should implement paste image filter from ITK
    for x in range( len( image_list ) ):
        if dimension == 2:
            newImage[:,:,x] = image_list[x][:,:]
        if dimension == 3:
            newImage[:,:,:,x] = image_list[x][:,:,:]
    return newImage


@image_method
def ndimage_to_list(image):
    """
    Split a n dimensional ANTsImage into a list
    of n-1 dimensional ANTsImages

    Arguments
    ---------
    image : ANTsImage
        n-dimensional image to split

    Returns
    -------
    list of ANTsImage types

    Example
    -------
    >>> import ants
    >>> image = ants.image_read(ants.get_ants_data('r16'))
    >>> image2 = ants.image_read(ants.get_ants_data('r16'))
    >>> imageTar = ants.make_image( ( *image2.shape, 2 ) )
    >>> image3 = ants.list_to_ndimage( imageTar, [image,image2])
    >>> image3.dimension == 3
    >>> images_unmerged = ants.ndimage_to_list( image3 )
    >>> len(images_unmerged) == 2
    >>> images_unmerged[0].dimension == 2
    """
    inpixeltype = image.pixeltype
    dimension = image.dimension
    components = 1
    imageShape = image.shape
    nSections = imageShape[ dimension - 1 ]
    subdimension = dimension - 1
    suborigin = ants.get_origin( image )[0:subdimension]
    subspacing = ants.get_spacing( image )[0:subdimension]
    subdirection = np.eye( subdimension )
    for i in range( subdimension ):
        subdirection[i,:] = ants.get_direction( image )[i,0:subdimension]
    subdim = image.shape[ 0:subdimension ]
    imagelist = []
    for i in range( nSections ):
        img = ants.slice_image( image, axis = subdimension, idx = i )
        ants.set_spacing( img, subspacing )
        ants.set_origin( img, suborigin )
        ants.set_direction( img, subdirection )
        imagelist.append( img )

    return imagelist




 

__all__ = ['smooth_image']

import math

from ants.decorators import image_method
from ants.internal import get_lib_fn

import ants


@image_method
def smooth_image(image, sigma, sigma_in_physical_coordinates=True, FWHM=False, max_kernel_width=32):
    """
    Smooth an image

    ANTsR function: `smoothImage`

    Arguments
    ---------
    image   
        Image to smooth
    
    sigma   
        Smoothing factor. Can be scalar, in which case the same sigma is applied to each dimension, or a vector of length dim(inimage) to specify a unique smoothness for each dimension.
    
    sigma_in_physical_coordinates : boolean  
        If true, the smoothing factor is in millimeters; if false, it is in pixels.
    
    FWHM : boolean    
        If true, sigma is interpreted as the full-width-half-max (FWHM) of the filter, not the sigma of a Gaussian kernel.
    
    max_kernel_width : scalar    
        Maximum kernel width
    
    Returns
    -------
    ANTsImage
    
    Example
    -------
    >>> import ants
    >>> image = ants.image_read( ants.get_ants_data('r16'))
    >>> simage = ants.smooth_image(image, (1.2,1.5))
    """
    if image.components == 1:
        outimage = image.clone()
        if not isinstance(sigma, (tuple,list)):
            sigma = [sigma]

        if isinstance(sigma, (tuple, list)) and ((len(sigma) != image.dimension) and (len(sigma) != 1)):
            raise ValueError('Length of sigma must be either 1 or the dimensionality of input image')

        image_float = image.clone('float')
        if FWHM:
            sigma = [s/2.355 for s in sigma]

        max_kernel_width = int(math.ceil(max_kernel_width))

        smooth_image_fn = get_lib_fn('SmoothImage')
        outimage = smooth_image_fn(image_float.pointer, sigma, sigma_in_physical_coordinates, max_kernel_width)
        ants_outimage = ants.from_pointer(outimage)
        return ants_outimage
    else:
        imagelist = ants.split_channels(image)
        newimages = []
        for image in imagelist:
            newimage = smooth_image(image, sigma, sigma_in_physical_coordinates, FWHM, max_kernel_width)
            newimages.append(newimage)
        return ants.merge_channels(newimages)

